"""
    Testing dataset read operations
"""

import pandas as pd
import time

from relevanceai import Client

from relevanceai.dataset import Dataset


def test_Dataset_init(test_dataset: Dataset):
    assert True


def test_Dataset_json_encoder(test_client: Client, obj_dataset_id: str):
    test_dataset = test_client.Dataset(obj_dataset_id)
    assert "value1" in test_dataset.schema


def test_info(test_dataset: Dataset):
    test_dataset.info()
    assert True


def test_shape(test_dataset: Dataset):
    test_dataset.shape
    assert True


def test_head(test_dataset: Dataset):
    test_dataset.head()
    assert True


def test_describe(test_dataset: Dataset):
    test_dataset.describe()
    assert True


def test_sample(test_dataset: Dataset):
    sample_n = test_dataset.sample(n=10)
    assert len(sample_n) == 10


def test_value_counts(test_dataset: Dataset):
    test_dataset["sample_1_label"].value_counts()
    assert True


def test_value_counts_normalize(test_dataset: Dataset):
    test_dataset["sample_1_label"].value_counts(normalize=True)
    assert True


def test_filter(test_dataset: Dataset):
    test_dataset.filter(items=["sample_1_label"])
    test_dataset.filter(like="sample_1_label")
    test_dataset.filter(regex="s$")
    assert True


def test_read_df_check(test_read_df, vector_documents):
    assert test_read_df is None, "Did not insert properly"


def test_datasets_schema(test_dataset: Dataset):
    test_dataset.schema
    assert True


def test_info(test_dataset: Dataset):
    info = test_dataset.info()
    assert isinstance(info, pd.DataFrame)


def test_df_get_smoke(test_dataset: Dataset):
    assert test_dataset.get(["321", "3421"])


def test_df_metadata(test_dataset: Dataset):
    metadata = {"value": "hey"}
    time.sleep(1)
    test_dataset.insert_metadata(metadata)
    time.sleep(1)
    new_metadata = test_dataset.metadata
    assert new_metadata["value"] == "hey"

    new_metadata = {"value": "cool", "old_value": "hey"}
    response = test_dataset.upsert_metadata(new_metadata)
    new_metadata = test_dataset.metadata
    assert new_metadata["value"] == "cool"
    assert new_metadata["old_value"] == "hey"



"""
Automated Cluster Reporting

.. warning::
    This is a beta feature.

.. note::
    **Introduced in v1.0.0.**

You can run cluster reporting as a standalone module with the option
to store it in Relevance AI.

.. code-block::

    import requests
    docs = requests.get("https://raw.githubusercontent.com/fanzeyi/pokemon.json/master/pokedex.json").json()
    for d in docs:
        b = d['base']
        d.update(b)
        d['base_vector_'] = [b["Attack"], b["Defense"], b["HP"], b["Sp. Attack"], b["Sp. Defense"], b["Speed"]]

    import pandas as pd
    import numpy as np
    df = pd.DataFrame(docs)
    X = np.array(df['base_vector_'].tolist())


    from relevanceai.workflows.cluster.reports.cluster_report import ClusterReport
    from sklearn.cluster import KMeans

    N_CLUSTERS = 2
    kmeans = KMeans(n_clusters=N_CLUSTERS)
    cluster_labels = kmeans.fit_predict(X)

    report = ClusterReport(
        X=X,
        cluster_labels=cluster_labels,
        num_clusters=N_CLUSTERS,
        model=kmeans
    )

    # JSON output
    report.internal_report

    # Prettyprinted report of overall statistics
    report.internal_overall_report

    # Storing your cluster report
    from relevanceai import Client
    client = Client()
    response = client.store_cluster_report(
        report_name="kmeans",
        report=report.internal_report
    )

    # Listing all cluster reports
    client.list_cluster_reports()

    # Deleting cluster report
    client.delete_cluster_report(response['_id])


You can also insert your own centroid vectors if you want them to be represented.
For example - you may want to measure off medoids (points in your dataset) instead of centroids
(as opposed to points outside of your dataset).

In the example below, we show how you calculate centroids or medoids for HDBSCAN

.. code-block::

    from relevanceai.workflows.cluster.reports.cluster_report import ClusterReport

    import hdbscan
    clusterer = hdbscan.HDBSCAN()

    cluster_labels = clusterer.fit_predict(X)
    centroids = ClusterReport.calculate_centroids(X, cluster_labels)
    # or if you want medoids
    # medoids = ClusterReport.calculate_medoids(X, cluster_labels)

    report = ClusterReport(
        X=X,
        cluster_labels=cluster_labels,
        centroid_vectors=centroids
    )
    report.internal_overall_report

.. code-block::

    from relevanceai.workflows.cluster.reports.cluster_report import ClusterReport

    import hdbscan
    clusterer = hdbscan.HDBSCAN()

    cluster_labels = clusterer.fit_predict(X)
    medoids = ClusterReport.calculate_medoids(X, cluster_labels)

    report = ClusterReport(
        X=X,
        cluster_labels=cluster_labels,
        centroid_vectors=centroids
    )
    report.internal_overall_report

"""

import warnings
import pandas as pd
import numpy as np
import functools

from relevanceai.constants.warning import Warning
from relevanceai.utils.integration_checks import (
    is_hdbscan_available,
    is_sklearn_available,
)
from relevanceai.reports.cluster.grading import get_silhouette_grade
from typing import Union, List, Dict, Any, Optional
from doc_utils import DocUtils
from relevanceai.utils.decorators.analytics import track_event_usage

if is_sklearn_available():
    from sklearn.metrics import (
        davies_bouldin_score,
        calinski_harabasz_score,
        silhouette_samples,
    )
    from sklearn.metrics.pairwise import (
        pairwise_distances,
    )
    from sklearn.cluster import MiniBatchKMeans, KMeans
    from sklearn.tree import _tree, DecisionTreeClassifier
    from sklearn.neighbors import NearestNeighbors

from relevanceai.constants.warning import Warning


class ClusterReport(DocUtils):
    """
    Receive an automated cluster reprot

    .. warning::
        This is a beta feature.

    .. note::
        **Introduced in v1.0.0.**


    Parameters
    -------------

    X: np.ndarray
        The original data
    cluster_labels: List[str]
        A list of cluster labels
    model
        The model to analyze. Currently only used
    num_clusters: Optional[int]
        The number of clusters. This is required if we can't actually tell how many clusters there are
    outlier_label: Optional[str, int]
        The label if it is an outlier
    centroids: Union[list, np.ndarray]
        The centroid vectors. If supplied, will use these. Otherwise, will try to infer them
        from the model.
    """

    @track_event_usage("cluster_report")
    def __init__(
        self,
        X: Union[list, np.ndarray],
        cluster_labels: Union[List[Union[str, float]], np.ndarray],
        model: KMeans = None,
        num_clusters: int = None,
        outlier_label: Union[str, int] = -1,
        centroids: Union[list, np.ndarray] = None,
        verbose: bool = False,
    ):

        self._typecheck_model(model)

        if isinstance(X, list):
            self.X = np.array(X)
        else:
            self.X = X
        if isinstance(cluster_labels, list):
            self.cluster_labels = np.array(cluster_labels)
        else:
            self.cluster_labels = cluster_labels
        self.num_clusters = (
            len(set(cluster_labels)) if num_clusters is None else num_clusters
        )
        self.model = model
        self.outlier_label = outlier_label
        self._typecheck_centroid_vectors(centroids)
        self._centroids = centroids
        self.verbose = verbose

    def _typecheck_centroid_vectors(
        self, centroid_vectors: Optional[Union[list, Dict, np.ndarray]] = None
    ):
        if isinstance(centroid_vectors, (list, np.ndarray)):
            warnings.warn(Warning.CENTROID_VECTORS)

    def _typecheck_model(self, model):
        # Warns users that their model may not be supported
        # this may not be necessary and is removable in future
        if is_hdbscan_available():
            if "hdbscan" in str(type(model)):
                return
        if is_sklearn_available():
            if "sklearn" in str(type(model)):
                return
        warnings.warn(Warning.MODEL_NOT_SUPPORTED)

    @staticmethod
    def summary_statistics(array: np.ndarray, axis=0):
        """
        Basic summary statistics
        """
        if axis == 2:
            return {
                "sum": array.sum(),
                "mean": array.mean(),
                "std": array.std(),
                "variance": array.var(),
                "min": array.min(),
                "max": array.max(),
                "12_5%": np.percentile(array, 12.5),
                "25%": np.percentile(array, 25),
                "37_5%": np.percentile(array, 37.5),
                "50%": np.percentile(array, 50),
                "62_5%": np.percentile(array, 62.5),
                "75%": np.percentile(array, 75),
                "87_5%": np.percentile(array, 87.5),
            }
        else:
            return {
                "sum": array.sum(axis=axis),
                "mean": array.mean(axis=axis),
                "std": array.std(axis=axis),
                "variance": array.var(axis=axis),
                "min": array.min(axis=axis),
                "max": array.max(axis=axis),
                "12_5%": np.percentile(array, 12.5, axis=axis),
                "25%": np.percentile(array, 25, axis=axis),
                "37_5%": np.percentile(array, 37.5, axis=axis),
                "50%": np.percentile(array, 50, axis=axis),
                "62_5%": np.percentile(array, 62.5, axis=axis),
                "75%": np.percentile(array, 75, axis=axis),
                "87_5%": np.percentile(array, 87.5, axis=axis),
            }

    def get_distance_from_centroid(self, cluster_data, center_vector):
        distances_from_centroid = pairwise_distances([center_vector], cluster_data)
        return ClusterReport.summary_statistics(distances_from_centroid, axis=2)

    def get_distance_from_centroid_to_another(self, other_cluster_data, center_vector):
        """Store the distances from a centroid to another."""
        distances_from_centroid_to_another = pairwise_distances(
            [center_vector], other_cluster_data
        )
        return ClusterReport.summary_statistics(
            distances_from_centroid_to_another, axis=2
        )

    def get_distance_from_grand_centroid(self, grand_centroid, specific_cluster_data):
        distances_from_grand_centroid = pairwise_distances(
            [grand_centroid], specific_cluster_data
        )
        return ClusterReport.summary_statistics(distances_from_grand_centroid, axis=2)

    def get_distance_from_grand_centroid_to_point_in_another_cluster(
        self, grand_centroid, other_cluster_data
    ):
        distances_from_grand_centroid_to_another = pairwise_distances(
            [grand_centroid], other_cluster_data
        )
        return ClusterReport.summary_statistics(
            distances_from_grand_centroid_to_another, axis=2
        )

    @staticmethod
    def get_z_score(value, mean, std):
        return (value - mean) / std

    @functools.lru_cache(maxsize=128)
    def get_centers(self, output_format="array"):
        if hasattr(self.model, "cluster_centers_"):
            return self.model.cluster_centers_
        elif hasattr(self.model, "get_centers"):
            return self.model.get_centers()
        elif self._centroids is not None:
            if output_format == "array":
                if isinstance(self._centroids, dict):
                    return np.array(list(self._centroids.values()))
                else:
                    return self._centroids
            else:
                return self._centroids
        else:
            if self.verbose:
                print(
                    "No centroids detected. We recommend including centroids to get all stats."
                )
            return

    @property  # type: ignore
    @functools.lru_cache(maxsize=128)
    def internal_report(self):
        """
        Provide the standard clustering report.
        """
        self.X_silhouette_scores = silhouette_samples(
            self.X, self.cluster_labels, metric="euclidean"
        )
        graded_score = self.X_silhouette_scores.mean()
        grade = get_silhouette_grade(graded_score)

        self._internal_report = {
            "grade": grade,
            "overall": {
                "summary": ClusterReport.summary_statistics(self.X),
                "davies_bouldin_score": davies_bouldin_score(
                    self.X, self.cluster_labels
                ),
                "calinski_harabasz_score": calinski_harabasz_score(
                    self.X, self.cluster_labels
                ),
                "silhouette_score": ClusterReport.summary_statistics(
                    self.X_silhouette_scores
                ),
            },
            "each": [
                # {
                #     "cluster_id": "cluster-1",
                #     "summary": {},
                #     "centers": {},
                #     "silhouette_score": {}
                # }
            ]
            # {
            #     "summary": {},
            #     "centers": {},
            #     "silhouette_score": {},
            # },
        }
        self._store_basic_centroid_stats(self._internal_report["overall"])

        labels, counts = np.unique(self.cluster_labels, return_counts=True)
        if self.verbose:
            print("Detected the cluster labels:")
            print(labels)

        cluster_report = {"frequency": {"total": 0, "each": {}}}

        for i, cluster_label in enumerate(labels):
            cluster_label_doc = {
                "cluster_id": str(cluster_label),
            }
            cluster_bool = self.cluster_labels == cluster_label

            specific_cluster_data = self.X[cluster_bool]
            other_cluster_data = self.X[~cluster_bool]

            cluster_frequency = counts[i]
            cluster_report["frequency"]["total"] += cluster_frequency
            cluster_report["frequency"]["each"][cluster_label] = cluster_frequency

            cluster_label_doc["summary"] = ClusterReport.summary_statistics(self.X)

            # If each value of the vector is important
            center_stats = {"by_features": {}}

            center_stats["by_features"]["summary"] = ClusterReport.summary_statistics(
                specific_cluster_data
            )

            if self.has_centers():

                grand_centroid = self.X[cluster_bool].mean(axis=0)

                centroid_vector = self._get_centroid_vector(
                    i, cluster_label, grand_centroid
                )

                squared_errors = np.square(
                    np.subtract(
                        [centroid_vector] * len(specific_cluster_data),
                        specific_cluster_data,
                    )
                )

                self._internal_report["overall"]["grand_centroids"].append(
                    grand_centroid
                )

                center_stats[
                    "distance_from_centroid"
                ] = self.get_distance_from_centroid(
                    specific_cluster_data, centroid_vector
                )

                center_stats[
                    "distance_from_centroid_to_point_in_another_cluster"
                ] = self.get_distance_from_centroid_to_another(
                    other_cluster_data, centroid_vector
                )

                center_stats["distances_from_grand_centroid"] = pairwise_distances(
                    [grand_centroid], specific_cluster_data
                )

                center_stats[
                    "distance_from_grand_centroid"
                ] = self.get_distance_from_grand_centroid(
                    grand_centroid, specific_cluster_data
                )

                center_stats[
                    "distance_from_grand_centroid_to_point_in_another_cluster"
                ] = self.get_distance_from_grand_centroid_to_point_in_another_cluster(
                    grand_centroid, other_cluster_data
                )

                center_stats["by_features"][
                    "overall_z_score"
                ] = ClusterReport.get_z_score(
                    centroid_vector,
                    self._internal_report["overall"]["summary"]["mean"],
                    self._internal_report["overall"]["summary"]["std"],
                )

                center_stats["by_features"]["z_score"] = ClusterReport.get_z_score(
                    centroid_vector,
                    cluster_label_doc["summary"]["mean"],
                    cluster_label_doc["summary"]["std"],
                )

                center_stats["by_features"][
                    "overall_z_score_grand_centroid"
                ] = ClusterReport.get_z_score(
                    grand_centroid,
                    self._internal_report["overall"]["summary"]["mean"],
                    self._internal_report["overall"]["summary"]["std"],
                )

                center_stats[
                    "overall_z_score_grand_centroid"
                ] = ClusterReport.summary_statistics(
                    center_stats["by_features"]["overall_z_score_grand_centroid"]
                )

                center_stats["by_features"]["z_score_grand_centroid"] = (
                    grand_centroid - cluster_label_doc["summary"]["mean"]
                ) / cluster_label_doc["summary"]["std"]

                # this might not be needed
                center_stats["overall_z_score"] = ClusterReport.summary_statistics(
                    center_stats["by_features"]["overall_z_score"]
                )

                # this might not be needed
                center_stats["z_score"] = ClusterReport.summary_statistics(
                    center_stats["by_features"]["z_score"]
                )
                center_stats[
                    "z_score_grand_centroid"
                ] = ClusterReport.summary_statistics(
                    center_stats["by_features"]["z_score_grand_centroid"]
                )

                # squared errors are calculted by the centroids
                center_stats["squared_errors"] = ClusterReport.summary_statistics(
                    squared_errors, axis=2
                )

                squared_errors_by_col = []

                for f in range(len(squared_errors[0])):
                    squared_errors_by_col.append(
                        {
                            "cluster_id": str(f),
                            "squared_errors": ClusterReport.summary_statistics(
                                squared_errors[:, f], axis=2
                            ),
                        }
                    )

                center_stats["by_features"] = squared_errors_by_col

                cluster_label_doc["centers"] = center_stats

            cluster_label_doc["silhouette_score"] = ClusterReport.summary_statistics(
                self.X_silhouette_scores[cluster_bool], axis=2
            )

            self._internal_report["each"].append(cluster_label_doc)

        if self.has_centers():

            min_centroid_distance = min(
                c["centers"]["distance_from_centroid"]["min"]
                for c in self._internal_report["each"]
            )

            max_centroid_distance = self._internal_report["overall"][
                "centroids_distance_matrix"
            ].max()

            self._internal_report["overall"]["dunn_index"] = self.dunn_index(
                min_centroid_distance, max_centroid_distance
            )

        return self._internal_report

    def _get_centroid_vector(
        self, i: int = None, cluster_label: int = None, default_vector=None
    ):
        # If this is an outlier, we will automatically default to the grand centroid
        if cluster_label == self.outlier_label:
            if self.verbose:
                print(
                    "Outlier labels detected. Using the grand centroid for the outlier label."
                )
            return default_vector
        centers = self.get_centers()
        if isinstance(centers, (list, np.ndarray)):
            centroid_vector = centers[i]  # type: ignore
        elif isinstance(centers, dict):
            try:
                centroid_vector = centers[cluster_label]
            except KeyError:
                warnings.warn(Warning.CLUSTER_LABEL_NOT_IN_CENTROIDS)
                centroid_vector = default_vector
        else:
            raise ValueError("Centroid vector needs to be a list or a dictionary.")
        return centroid_vector

    def dunn_index(self, min_distance_from_centroid, max_centroid_distance):
        return min_distance_from_centroid / max_centroid_distance

    def has_centers(self):
        return self.get_centers() is not None

    def _store_basic_centroid_stats(self, overall_report):
        """Store"""
        if self.has_centers():
            centroids = self.get_centers()
            overall_report["centroids"] = centroids
            overall_report["centroids_distance_matrix"] = pairwise_distances(
                centroids, metric="euclidean"
            )
            overall_report["grand_centroids"] = []
            overall_report["average_distance_between_centroids"] = (
                overall_report["centroids_distance_matrix"].sum(axis=1) - 1
            ) / self.num_clusters

    @property
    def davies_bouldin_score(self):
        return pd.DataFrame(
            self.subset_documents(
                ["davies_bouldin_score"], [self.internal_report["overall"]]
            )
        )

    @property
    def calinski_harabasz_score(self):
        return pd.DataFrame(
            self.subset_documents(
                ["calinski_harabasz_score"], [self.internal_report["overall"]]
            )
        )

    @property
    def internal_overall_report(self):
        """
        View the internal overall report.
        """
        # TODO: Figure out a better way to present this than having index in the middle
        if not self.has_centers():
            metrics = pd.DataFrame(
                self.subset_documents(
                    ["davies_bouldin_score", "calinski_harabasz_score"],
                    [self.internal_report["overall"]],
                )
            )
            overall_df = pd.DataFrame(self.internal_report["overall"])[
                ["summary", "silhouette_score"]
            ]
            return pd.concat([metrics, overall_df.reset_index()], axis=1).fillna(" ")
        else:
            metrics = pd.DataFrame(
                self.subset_documents(
                    ["davies_bouldin_score", "calinski_harabasz_score", "dunn_index"],
                    [self.internal_report["overall"]],
                )
            )

            overall_df = pd.DataFrame(
                self.get_fields(
                    ["overall.summary", "overall.silhouette_score"],
                    self.internal_report,
                )
            ).T
            overall_df.columns = ["summary", "silhouette_score"]
            return pd.concat([metrics, overall_df.reset_index()], axis=1).fillna(" ")

    @staticmethod
    def calculate_centroids(X, cluster_labels):
        """Calculate the centes"""
        centroid_vectors = {}
        for label in cluster_labels:
            centroid_vectors[label] = X[cluster_labels == label].mean(axis=0)
        return centroid_vectors

    @staticmethod
    def calculate_medoids(X, cluster_labels):
        centroids = ClusterReport.calculate_centroids(X, cluster_labels)
        medoids = {}
        nbrs = NearestNeighbors(n_neighbors=1, algorithm="ball_tree").fit(X)
        medoid_indexes = nbrs.kneighbors(
            np.array(list(centroids.values())), n_neighbors=1, return_distance=False
        )
        medoids = X[medoid_indexes]
        return medoids

    def get_class_rules(self, tree: DecisionTreeClassifier, feature_names: list):
        self.inner_tree: _tree.Tree = tree.tree_
        self.classes = tree.classes_
        self.class_rules_dict: Dict[Any, Any] = dict()
        self.tree_dfs()

    def tree_dfs(self, node_id=0, current_rule: Optional[list] = None):
        current_rule = [] if current_rule is None else current_rule
        # if not hasattr(self, "classes"):
        #    self.get_class_rules()

        # feature[i] holds the feature to split on, for the internal node i.
        split_feature = self.inner_tree.feature[node_id]
        if split_feature != _tree.TREE_UNDEFINED:  # internal node
            name = self.feature_names[split_feature]
            threshold = self.inner_tree.threshold[node_id]
            # left child
            left_rule = current_rule + ["({} <= {})".format(name, threshold)]
            self.tree_dfs(self.inner_tree.children_left[node_id], left_rule)
            # right child
            right_rule = current_rule + ["({} > {})".format(name, threshold)]
            self.tree_dfs(self.inner_tree.children_right[node_id], right_rule)
        else:  # leaf
            dist = self.inner_tree.value[node_id][0]
            dist = dist / dist.sum()
            max_idx = dist.argmax()
            if len(current_rule) == 0:
                rule_string = "ALL"
            else:
                rule_string = " and ".join(current_rule)
            # register new rule to dictionary
            selected_class = self.classes[max_idx]
            class_probability = dist[max_idx]
            class_rules = self.class_rules_dict.get(selected_class, [])
            class_rules.append((rule_string, class_probability))
            self.class_rules_dict[selected_class] = class_rules

    def cluster_reporting(self, data: pd.DataFrame, clusters, max_depth: int = 5):
        # Create Model
        tree = DecisionTreeClassifier(max_depth=max_depth, criterion="entropy")
        tree.fit(data, clusters)
        print(tree.score(data, clusters))

        # Generate Report
        self.feature_names = data.columns
        self.get_class_rules(tree, self.feature_names)

        report_class_list = []

        for class_name in self.class_rules_dict.keys():
            rule_list = self.class_rules_dict[class_name]
            combined_string = ""
            for rule in rule_list:
                combined_string += "[{}] {}\n\n".format(rule[1], rule[0])
            report_class_list.append((class_name, combined_string))

        cluster_instance_df = pd.Series(clusters).value_counts().reset_index()
        cluster_instance_df.columns = ["class_name", "instance_count"]

        report_df = pd.DataFrame(report_class_list, columns=["class_name", "rule_list"])
        report_df = pd.merge(
            cluster_instance_df, report_df, on="class_name", how="left"
        )
        return report_df.sort_values(by="class_name")[
            ["class_name", "instance_count", "rule_list"]
        ]



from relevanceai.recipes.sentiment.sentiments import SentimentOps
from relevanceai.recipes.base import Workflow


class SentimentWorkflow(Workflow, SentimentOps):
    def fit_dataset(
        self,
        dataset,
        input_field: str,
        output_field: str,
        log_to_file: bool = True,
        chunksize: int = 20,
        workflow_alias: str = "sentiment",
        notes=None,
    ):
        workflow = Workflow(
            self.analyze_sentiment, workflow_alias=workflow_alias, notes=notes
        )
        workflow.fit_dataset(
            dataset=dataset,
            input_field=input_field,
            output_field=output_field,
            log_to_file=log_to_file,
            chunksize=chunksize,
        )



from tests.globals.document.error_document import *
from tests.globals.document.simple_document import *
from tests.globals.document.datetime_document import *
from tests.globals.document.dataclass_document import *
from tests.globals.document.nested_document import *
from tests.globals.document.numpy_document import *
from tests.globals.document.pandas_document import *
from tests.globals.document.vector_document import *



# -*- coding: utf-8 -*-
"""Batch Retrieve"""

import math
import traceback

from typing import List, Optional

from relevanceai._api.batch.chunk import Chunker
from relevanceai._api.endpoints.api_client import APIEndpointsClient

from relevanceai.utils.cache import lru_cache
from relevanceai.utils.progress_bar import progress_bar

from relevanceai.constants.constants import MAX_CACHESIZE

# ADD SUPPORT FOR SAVING TO JSON


class BatchRetrieveClient(APIEndpointsClient, Chunker):
    def _get_documents(
        self,
        dataset_id: str,
        number_of_documents: int = 20,
        filters: Optional[list] = None,
        cursor: str = None,
        batch_size: int = 1000,
        sort: Optional[list] = None,
        select_fields: Optional[list] = None,
        include_vector: bool = True,
        include_cursor: bool = False,
    ):
        """
        Retrieve documents with filters. Filter is used to retrieve documents that match the conditions set in a filter query. This is used in advance search to filter the documents that are searched. \n
        If you are looking to combine your filters with multiple ORs, simply add the following inside the query {"strict":"must_or"}.
        Parameters
        ----------
        dataset_id: string
            Unique name of dataset
        number_of_documents: int
            Number of documents to retrieve
        select_fields: list
            Fields to include in the search results, empty array/list means all fields.
        cursor: string
            Cursor to paginate the document retrieval
        batch_size: int
            Number of documents to retrieve per iteration
        include_vector: bool
            Include vectors in the search results
        sort: list
            Fields to sort by. For each field, sort by descending or ascending. If you are using descending by datetime, it will get the most recent ones.
        filters: list
            Query for filtering the search results
        """
        filters = [] if filters is None else filters
        sort = [] if sort is None else sort
        select_fields = [] if select_fields is None else select_fields

        if batch_size > number_of_documents:
            batch_size = number_of_documents

        resp = self.datasets.documents.get_where(
            dataset_id=dataset_id,
            select_fields=select_fields,
            include_vector=include_vector,
            page_size=batch_size,
            sort=sort,
            is_random=False,
            random_state=0,
            filters=filters,
            cursor=cursor,
        )
        data = resp["documents"]

        if number_of_documents > batch_size:
            _cursor = resp["cursor"]
            _page = 0
            while resp:
                self.logger.debug(f"Paginating {_page} batch size {batch_size} ...")
                resp = self.datasets.documents.get_where(
                    dataset_id=dataset_id,
                    select_fields=select_fields,
                    include_vector=include_vector,
                    page_size=batch_size,
                    sort=sort,
                    is_random=False,
                    random_state=0,
                    filters=filters,
                    cursor=_cursor,
                )
                _data = resp["documents"]
                _cursor = resp["cursor"]
                if (_data == []) or (_cursor == []):
                    break
                data += _data
                if number_of_documents and (len(data) >= int(number_of_documents)):
                    break
                _page += 1
            data = data[:number_of_documents]

        if include_cursor:
            return {"documents": data, "cursor": resp["cursor"]}
        return data

    @lru_cache(maxsize=MAX_CACHESIZE)
    def _get_all_documents(
        self,
        dataset_id: str,
        chunksize: int = 1000,
        filters: Optional[List] = None,
        sort: Optional[List] = None,
        select_fields: Optional[List] = None,
        include_vector: bool = True,
        show_progress_bar: bool = True,
    ):
        """
        Retrieve all documents with filters. Filter is used to retrieve documents that match the conditions set in a filter query. This is used in advance search to filter the documents that are searched. For more details see documents.get_where.

        Example
        ---------

        >>> client = Client()
        >>> client.get_all_documents(dataset_id="sample_dataset_id"")

        Parameters
        ----------
        dataset_id : string
            Unique name of dataset
        chunksize: list
            Number of documents to retrieve per retrieval
        include_vector: bool
            Include vectors in the search results
        sort: list
            Fields to sort by. For each field, sort by descending or ascending. If you are using descending by datetime, it will get the most recent ones.
        filters: list
            Query for filtering the search results
        select_fields : list
            Fields to include in the search results, empty array/list means all fields.
        """
        filters = [] if filters is None else filters
        sort = [] if sort is None else sort
        select_fields = [] if select_fields is None else select_fields

        # Initialise values
        length = 1
        cursor = None
        full_data = []

        # Find number of iterations
        number_of_documents = self.get_number_of_documents(
            dataset_id=dataset_id, filters=filters
        )
        iterations_required = math.ceil(number_of_documents / chunksize)

        # While there is still data to fetch, fetch it at the latest cursor

        for _ in progress_bar(
            range(iterations_required), show_progress_bar=show_progress_bar
        ):
            x = self.datasets.documents.get_where(
                dataset_id,
                filters=filters,
                cursor=cursor,
                page_size=chunksize,
                sort=sort,
                select_fields=select_fields,
                include_vector=include_vector,
            )
            try:
                length = len(x["documents"])
                cursor = x["cursor"]

                # Append fetched data to the full data
                if length > 0:
                    full_data += x["documents"]
            except Exception as e:
                traceback.print_exc()
                pass
        return full_data

    def get_number_of_documents(self, dataset_id, filters: Optional[List] = None):
        """
        Get number of documents in a dataset. Filter can be used to select documents that match the conditions set in a filter query. For more details see documents.get_where.

        Parameters
        ----------
        dataset_ids: list
            Unique names of datasets
        filters: list
            Filters to select documents
        """
        filters = [] if filters is None else filters

        return self.datasets.documents.get_where(
            dataset_id, page_size=1, filters=filters
        )["count"]



from tests.globals.datasets.datetime_datasets import *
from tests.globals.datasets.large_datasets import *
from tests.globals.datasets.misc_datasets import *
from tests.globals.datasets.nested_datasets import *
from tests.globals.datasets.numpy_datasets import *
from tests.globals.datasets.pandas_datasets import *
from tests.globals.datasets.sample_datasets import *
from tests.globals.datasets.vector_datasets import *



SEARCH_APP_LINK = "https://cloud.relevance.ai/dataset/{}/deploy/recent/search"
PROJECTOR_APP_LINK = "https://cloud.relevance.ai/dataset/{}/deploy/recent/projector"
CLUSTER_APP_LINK = "https://cloud.relevance.ai/dataset/{}/deploy/recent/cluster/"

SIGNUP_URL = "https://cloud.relevance.ai/sdk/api"
AUSTRALIA_URL = "https://gateway-api-aueast.relevance.ai/latest"
WIDER_URL = "https://api.{}.relevance.ai/latest"
DEV_URL = "https://api-dev.{}.relevance.ai/latest"



"""
Modified Python code from:
https://github.com/python/cpython/blob/main/Lib/functools.py#L511

Key purposes of this new caching:
- Built-in support for hashing lists
- Built-in support for hashing dictionaries
"""

from threading import RLock
from functools import update_wrapper
from collections import namedtuple
from urllib.parse import MAX_CACHE_SIZE
from relevanceai.constants.constants import MAX_CACHESIZE

_CacheInfo = namedtuple("_CacheInfo", ["hits", "misses", "maxsize", "currsize"])


def string_hash(obj):
    return hash(str(obj))


class _HashedSeq(list):
    """This class guarantees that hash() will be called no more than once
    per element.  This is important because the lru_cache() will hash
    the key multiple times on a cache miss.
    """

    __slots__ = "hashvalue"

    def __init__(self, tup, hash=string_hash):
        self[:] = tup
        self.hashvalue = hash(tup)

    def __hash__(self):
        return self.hashvalue


def _make_key(
    args,
    kwds,
    typed,
    kwd_mark=(object(),),
    fasttypes={int, str},
    tuple=tuple,
    type=type,
    len=len,
):
    """Make a cache key from optionally typed positional and keyword arguments

    The key is constructed in a way that is flat as possible rather than
    as a nested structure that would take more memory.

    If there is only a single argument and its data type is known to cache
    its hash value, then that argument is returned without a wrapper.  This
    saves space and improves lookup speed.

    """
    # All of code below relies on kwds preserving the order input by the user.
    # Formerly, we sorted() the kwds before looping.  The new way is *much*
    # faster; however, it means that f(x=1, y=2) will now be treated as a
    # distinct call from f(y=2, x=1) which will be cached separately.
    key = args
    if kwds:
        key += kwd_mark
        for item in kwds.items():
            key += item
    if typed:
        key += tuple(type(v) for v in args)
        if kwds:
            key += tuple(type(v) for v in kwds.values())
    elif len(key) == 1 and type(key[0]) in fasttypes:
        return key[0]
    return _HashedSeq(key)


def _lru_cache_wrapper(user_function, maxsize, typed, _CacheInfo):
    # Constants shared by all lru cache instances:
    sentinel = object()  # unique object used to signal cache misses
    make_key = _make_key  # build a key from the function arguments
    PREV, NEXT, KEY, RESULT = 0, 1, 2, 3  # names for the link fields

    cache = {}
    hits = misses = 0
    full = False
    cache_get = cache.get  # bound method to lookup a key or return None
    cache_len = cache.__len__  # get cache size without calling len()
    lock = RLock()  # because linkedlist updates aren't threadsafe
    root = []  # root of the circular doubly linked list
    root[:] = [root, root, None, None]  # initialize by pointing to self

    if maxsize == 0:

        def wrapper(*args, **kwds):
            # No caching -- just a statistics update
            nonlocal misses
            misses += 1
            result = user_function(*args, **kwds)
            return result

    elif maxsize is None:

        def wrapper(*args, **kwds):
            # Simple caching without ordering or size limit
            nonlocal hits, misses
            key = make_key(args, kwds, typed)
            result = cache_get(key, sentinel)
            if result is not sentinel:
                hits += 1
                return result
            misses += 1
            result = user_function(*args, **kwds)
            cache[key] = result
            return result

    else:

        def wrapper(*args, **kwds):
            # Size limited caching that tracks accesses by recency
            nonlocal root, hits, misses, full
            key = make_key(args, kwds, typed)
            with lock:
                link = cache_get(key)
                if link is not None:
                    # Move the link to the front of the circular queue
                    link_prev, link_next, _key, result = link
                    link_prev[NEXT] = link_next
                    link_next[PREV] = link_prev
                    last = root[PREV]
                    last[NEXT] = root[PREV] = link
                    link[PREV] = last
                    link[NEXT] = root
                    hits += 1
                    return result
                misses += 1
            result = user_function(*args, **kwds)
            with lock:
                if key in cache:
                    # Getting here means that this same key was added to the
                    # cache while the lock was released.  Since the link
                    # update is already done, we need only return the
                    # computed result and update the count of misses.
                    pass
                elif full:
                    # Use the old root to store the new key and result.
                    oldroot = root
                    oldroot[KEY] = key
                    oldroot[RESULT] = result
                    # Empty the oldest link and make it the new root.
                    # Keep a reference to the old key and old result to
                    # prevent their ref counts from going to zero during the
                    # update. That will prevent potentially arbitrary object
                    # clean-up code (i.e. __del__) from running while we're
                    # still adjusting the links.
                    root = oldroot[NEXT]
                    oldkey = root[KEY]
                    oldresult = root[RESULT]
                    root[KEY] = root[RESULT] = None
                    # Now update the cache dictionary.
                    del cache[oldkey]
                    # Save the potentially reentrant cache[key] assignment
                    # for last, after the root and links have been put in
                    # a consistent state.
                    cache[key] = oldroot
                else:
                    # Put result in a new link at the front of the queue.
                    last = root[PREV]
                    link = [last, root, key, result]
                    last[NEXT] = root[PREV] = cache[key] = link
                    # Use the cache_len bound method instead of the len() function
                    # which could potentially be wrapped in an lru_cache itself.
                    full = cache_len() >= maxsize
            return result

    def cache_info():
        """Report cache statistics"""
        with lock:
            return _CacheInfo(hits, misses, maxsize, cache_len())

    def cache_clear():
        """Clear the cache and cache statistics"""
        nonlocal hits, misses, full
        with lock:
            cache.clear()
            root[:] = [root, root, None, None]
            hits = misses = 0
            full = False

    wrapper.cache_info = cache_info
    wrapper.cache_clear = cache_clear
    return wrapper


def lru_cache(maxsize=MAX_CACHE_SIZE, typed=False):
    """Least-recently-used cache decorator.
    If *maxsize* is set to None, the LRU features are disabled and the cache
    can grow without bound.
    If *typed* is True, arguments of different types will be cached separately.
    For example, f(3.0) and f(3) will be treated as distinct calls with
    distinct results.
    Arguments to the cached function must be hashable.
    View the cache statistics named tuple (hits, misses, maxsize, currsize)
    with f.cache_info().  Clear the cache and statistics with f.cache_clear().
    Access the underlying function with f.__wrapped__.
    See:  https://en.wikipedia.org/wiki/Cache_replacement_policies#Least_recently_used_(LRU)
    """

    # Users should only access the lru_cache through its public API:
    #       cache_info, cache_clear, and f.__wrapped__
    # The internals of the lru_cache are encapsulated for thread safety and
    # to allow the implementation to change (including a possible C version).

    if isinstance(maxsize, int):
        # Negative maxsize is treated as 0
        if maxsize < 0:
            maxsize = 0
    elif callable(maxsize) and isinstance(typed, bool):
        # The user_function was passed in directly via the maxsize argument
        user_function, maxsize = maxsize, 128
        wrapper = _lru_cache_wrapper(user_function, maxsize, typed, _CacheInfo)
        wrapper.cache_parameters = lambda: {"maxsize": maxsize, "typed": typed}
        return update_wrapper(wrapper, user_function)
    elif maxsize is not None:
        raise TypeError("Expected first argument to be an integer, a callable, or None")

    def decorating_function(user_function):
        wrapper = _lru_cache_wrapper(user_function, maxsize, typed, _CacheInfo)
        wrapper.cache_parameters = lambda: {"maxsize": maxsize, "typed": typed}
        return update_wrapper(wrapper, user_function)

    return decorating_function



import pandas as pd

from relevanceai.utils.cache import lru_cache

from relevanceai.dataset.read import Read
from relevanceai.utils.decorators.version import added
from relevanceai.constants.constants import MAX_CACHESIZE


class PandasExport(Read):
    @added(version="1.1.5")
    @lru_cache(maxsize=MAX_CACHESIZE)
    def to_pandas_dataframe(self, **kwargs) -> pd.DataFrame:
        """
        Converts a Relevance AI Dataset to a pandas DataFrame.

        Parameters
        ----------
        kwargs: Optional
            see client.get_all_documents() for extra args

        Example
        -------
        .. code-block::
            from relevanceai import Client

            client = Client()

            relevanceai_dataset = client.Dataset("dataset_id")
            df = relevance_ai.to_pandas_dataframe()
        """
        documents = self.get_all_documents(**kwargs)

        try:
            df = pd.DataFrame(documents)
            df.set_index("_id", inplace=True)
            return df
        except KeyError:
            raise Exception("No documents found")



from functools import wraps


def list_to_tuple(function):
    @wraps(function)
    def wrapper(*args, **kw):
        args = [tuple(x) if type(x) == list else x for x in args]
        new_kw = {}
        for k, v in kw.items():
            if type(v) == list:
                new_kw[k] = tuple(v)
            else:
                new_kw[k] = v
        result = function(*args, **new_kw)
        result = tuple(result) if type(result) == list else result
        return result

    return wrapper



# -*- coding: utf-8 -*-
"""
Test the clustering workflow from getting the documents, clustering and then inserting the relevant centroids
"""

import pytest

import time
from relevanceai import Client
from relevanceai.dataset import Dataset

from relevanceai.operations.cluster import ClusterOps
from relevanceai.operations.cluster.base import CentroidClusterBase


@pytest.mark.skip(reason="ClusterOps fit method missing")
def test_dataset_api_kmeans_integration(test_client: Client, test_dataset: Dataset):
    from sklearn.cluster import KMeans

    vector_field = "sample_1_vector_"
    alias = "test_alias"

    clusterer = test_client.ClusterOps(model=KMeans(n_clusters=2), alias=alias)

    clusterer.fit(dataset=test_df, vector_fields=[vector_field])

    assert f"_cluster_.{vector_field}.{alias}" in test_dataset.schema



import threading


def fire_and_forget(f):
    """
    Use as such:

    Example
    ----------

    .. code-block::

        @fire_and_forget
        def send_analytics():
            ...
    """

    def wrapped():
        threading.Thread(target=f).start()

    return wrapped



"""
Migrate from mongo database to Relevance Ai:

.. code-block::

    from relevanceai.api.batch import MongoImporter

    # Create an object of MongoImporter class
    connection_string= "..."
    project= "..."
    api_key= "..."
    mongo_importer = MongoImporter(connection_string, project, api_key)

    # Get a summary of the mondo database using "mongo_summary"
    mongo_importer.mongo_summary()

    # Set the desired source mongo collection using "set_mongo_collection"
    db_name = '...'
    collection_name = '...'
    mongo_importer.set_mongo_collection(db_name, dataset_id)

    # Get total number of entries in the mongo collection using "mongo_document_count"
    document_count = mongo_importer.mongo_document_count()

    # Migrate data from mongo to Relevance AI using "migrate_mongo2relevance_ai"
    chunk_size = 5000      # migrate batches of 5000 (default 2000)
    start_idx= 12000       # loads from mongo starting at index 12000 (default 0)
    dataset_id = "..."     # dataset id in the Relevance Ai platform
    mongo_importer.migrate(
        dataset_id, document_count, chunk_size=chunk_size,
        start_idx=start_idx)

"""

import copy
import json
import math
import numpy as np
import pandas as pd
import warnings

from tqdm.auto import tqdm
from typing import List

from relevanceai.utils import make_id
from relevanceai.constants.warning import Warning

try:
    from relevanceai import Client
    from pymongo import MongoClient

    PYMONGO_AVAILABLE = True
except (ImportError, ModuleNotFoundError):
    PYMONGO_AVAILABLE = False
    warnings.warn(Warning.MISSING_PACKAGE)

try:
    from bson import json_util

    BSON_AVAILABLE = True
except (ImportError, ModuleNotFoundError):
    BSON_AVAILABLE = False
    warnings.warn(Warning.MISSING_PACKAGE)


class MongoImporter(Client):
    def __init__(self, connection_string: str):
        super().__init__()
        if PYMONGO_AVAILABLE:
            self.mongo_client = MongoClient(connection_string)
        else:
            self.logger.error(
                "you are missing Pymongo. Please install this using `pip install pymongo==3.12`"
            )

    def mongo_summary(self):
        """
        returns a dictionary {key:value}
        key = db names
        value = collection names in each db
        """
        summary = {}
        for db in self.mongo_client.database_names():
            summary[db] = []
            for collection in self.mongo_client[db].collection_names():
                summary[db].append(collection)
        return summary

    def get_mongo_db(self, db_name: str):
        return self.mongo_client[db_name]

    def get_mongo_collection(self, db_name: str, collection_name: str):
        return self.mongo_client[db_name][collection_name]

    def set_mongo_db(self, db_name: str):
        self.mongodb = self.mongo_client[db_name]

    def set_mongo_collection(self, db_name: str, collection_name: str):
        self.mongo_collection = self.mongo_client[db_name][collection_name]

    def mongo_document_count(self):
        return self.mongo_collection.count()

    def create_relevance_ai_dataset(self, dataset_id: str):
        response = self.datasets.create(dataset_id)
        return response

    def update_id(self, documents: List[dict]):
        # makes bson id format json campatible
        for document in documents:
            try:
                document["_id"] = document["_id"]["$oid"]
            except Exception as e:
                self.logger.info("Could not use the original id: " + str(e))
                document["_id"] = make_id(document)
        return documents

    @staticmethod
    def parse_json(data):
        return json.loads(json_util.dumps(data))

    @staticmethod
    def flatten_inner_indxs(documents: List[dict]):
        # {f1:{f2:v}} => {f1-f2:v}
        expanded = copy.deepcopy(documents)
        for i, doc in enumerate(documents):
            for f, v in doc.items():
                if isinstance(v, dict):
                    del expanded[i][f]
                    for k in v:
                        expanded[i][f + "-" + k] = v[k]
        return expanded

    @staticmethod
    def remove_nan(documents: List[dict], replace_with: str = ""):
        for doc in documents:
            for f, v in doc.items():
                if isinstance(v, float) and math.isnan(v) or v == np.NaN:
                    doc[f] = replace_with
        return documents

    @staticmethod
    def build_range(document_count: int, chunk_size: int = 2000, start_idx: int = 0):
        rng = [
            (s, s + chunk_size)
            if s + chunk_size <= start_idx + document_count
            else (s, start_idx + document_count)
            for s in list(range(start_idx, start_idx + document_count, chunk_size))
        ]
        return rng

    def fetch_mongo_collection_data(self, start_idx: int = None, end_idx: int = None):
        if start_idx and end_idx:
            return list(self.mongo_collection.find()[start_idx:end_idx])
        return list(self.mongo_collection.find())

    def migrate(
        self,
        dataset_id: str,
        document_count: int,
        chunk_size: int = 2000,
        start_idx: int = 0,
        overwite: bool = False,
    ):
        """
        Migrate your MongoDB dataset ID.

        Parameters
        ------------
        dataset_id: str
            Name of your dataset
        document_count: int
            The number of documents in your collection
        chunk_size: int
            The number of chunks
        start_idx: int
            The start index in case it breaks
        overwrite: bool
            If True, then the dataset ID in Relevance AI will be overwritten

        """
        response = self.create_relevance_ai_dataset(dataset_id)
        if "already exists" in response["message"] and not overwite:
            self.logger.error(response["message"])
            return response["message"]

        total_ingest_cnt = 0
        for s_idx, e_idx in tqdm(
            MongoImporter.build_range(document_count, chunk_size, start_idx)
        ):
            df = pd.DataFrame(self.fetch_mongo_collection_data(s_idx, e_idx))
            documents = self.update_id(MongoImporter.parse_json(df.to_dict("records")))
            documents = MongoImporter.remove_nan(
                MongoImporter.flatten_inner_indxs(documents)
            )
            self.insert_documents(dataset_id, documents)
            total_ingest_cnt += len(documents)

        self.logger.info(
            f"Successfully ingested {total_ingest_cnt} entities to {dataset_id}."
        )
        return f"Successfully ingested {total_ingest_cnt} entities to {dataset_id}."



import random

from relevanceai.utils import make_id

from tests.globals.constants import generate_random_vector


def simple_document():
    document = {
        "value": random.randint(0, 1000),
        "sample_1_vector_": generate_random_vector(),
    }
    document["_id"] = make_id(document)
    return document



"""Tasks Module
"""
import time

from relevanceai.client.helpers import Credentials
from relevanceai.utils.base import _Base


class TasksClient(_Base):
    def __init__(self, credentials: Credentials):
        super().__init__(credentials)

    def create(self, dataset_id, task_name, task_parameters):
        """
        Tasks unlock the power of VecDb AI by adding a lot more new functionality with a flexible way of searching.

        Parameters
        ----------
        dataset_id : string
            Unique name of dataset
        task_name : string
            Name of task to complete
        task_parameters: dict
            Parameters of task to complete
        """
        return self.make_http_request(
            endpoint=f"/datasets/{dataset_id}/tasks/create",
            method="POST",
            parameters={"task_name": task_name, **task_parameters},
        )

    def status(self, dataset_id: str, task_id: str):
        """
        Get status of a collection level job. Whether its starting, running, failed or finished.

        Parameters
        ----------
        dataset_id : string
            Unique name of dataset
        task_id : string
            Unique name of task
        """
        return self.make_http_request(
            endpoint=f"/datasets/{dataset_id}/tasks/{task_id}/status", method="GET"
        )

    def list(self, dataset_id: str, show_active_only: bool = True):
        """
        List and get a history of all the jobs and its job_id, parameters, start time, etc.

        Parameters
        ----------
        dataset_id : string
            Unique name of dataset
        show_active_only : bool
            Whether to show active only
        """
        return self.make_http_request(
            endpoint=f"/datasets/{dataset_id}/tasks/list",
            method="GET",
            parameters={
                "show_active_only": show_active_only,
            },
        )

    def _loop_status_until_finish(
        self,
        dataset_id: str,
        task_id: str,
        verbose: bool = True,
        time_between_ping: int = 10,
    ):
        status = False

        while status not in ["Finished", "success"]:
            time.sleep(time_between_ping)
            try:
                status = self.status(dataset_id, task_id)["status"]
            except:
                self.logger.error(f"Status-check timed out: {task_id}")
                return task_id

            if verbose == True:
                self.logger.info(status)

        self.logger.success(f"Your task is {status}!")
        return

    def _check_status_until_finish(
        self,
        dataset_id: str,
        task_id: str,
        status_checker: bool = True,
        verbose: bool = True,
        time_between_ping: int = 10,
    ):

        if status_checker == True:
            self.logger.info(f"Task_ID: {task_id}")
            self._loop_status_until_finish(
                dataset_id,
                task_id,
                verbose=verbose,
                time_between_ping=time_between_ping,
            )
            return

        else:
            self.logger.info(
                "To view the progress of your job, visit https://cloud.relevanceai.com/collections/dashboard/jobs"
            )
            return {"task_id": task_id}

    # Note: The following tasks are instantiated manually to accelerate
    # creation of certain popular tasks

    # Make decorator wrap for all task checkers

    def create_cluster_task(
        self,
        dataset_id,
        vector_field: str,
        n_clusters: int,
        alias: str = "default",
        refresh: bool = False,
        n_iter: int = 10,
        n_init: int = 5,
        status_checker: bool = True,
        verbose: bool = True,
        time_between_ping: int = 10,
    ):

        """
        Start a task which creates clusters for a dataset based on a vector field
        Parameters
        ----------
        dataset_id : string
            Unique name of dataset
        vector_field : string
            The field to cluster on.
        alias: string
            Alias is used to name a cluster
        n_clusters: int
            Number of clusters to be specified.
        n_iter: int
            Number of iterations in each run
        n_init: int
            Number of runs to run with different centroid seeds
        refresh: bool
            Whether to rerun task on the whole dataset or just the ones missing the output
        """

        task = self.make_http_request(
            endpoint=f"/datasets/{dataset_id}/tasks/create",
            method="POST",
            parameters={
                "task_name": "ClusterOps",
                "vector_field": vector_field,
                "n_clusters": n_clusters,
                "alias": alias,
                "refresh": refresh,
                "n_iter": n_iter,
                "n_init": n_init,
            },
        )

        output = self._check_status_until_finish(
            dataset_id,
            task["task_id"],
            status_checker=status_checker,
            verbose=verbose,
            time_between_ping=time_between_ping,
        )
        return output

    def create_numeric_encoder_task(
        self,
        dataset_id: str,
        fields: list,
        vector_name: str = "_vector_",
        status_checker: bool = True,
        verbose: bool = True,
        time_between_ping: int = 10,
    ):
        """
        Within a collection encode the specified dictionary field in every document into vectors. \n
        For example: a dictionary that represents a person's characteristics visiting a store:
        >>> document 1 field: {"person_characteristics" : {"height":180, "age":40, "weight":70}}
        >>> document 2 field: {"person_characteristics" : {"age":32, "purchases":10, "visits": 24}}
        >>> -> <Encode the dictionaries to vectors> ->
        >>> | height | age | weight | purchases | visits |
        >>> |--------|-----|--------|-----------|--------|
        >>> | 180    | 40  | 70     | 0         | 0      |
        >>> | 0      | 32  | 0      | 10        | 24     |
        >>> document 1 dictionary vector: {"person_characteristics_vector_": [180, 40, 70, 0, 0]}
        >>> document 2 dictionary vector: {"person_characteristics_vector_": [0, 32, 0, 10, 24]}
        Parameters
        ----------
        dataset_id : string
            Unique name of dataset
        fields : list
            The numeric fields to encode into vectors.
        vector_name: string
            The name of the vector field created
        """
        task = self.make_http_request(
            endpoint=f"/datasets/{dataset_id}/tasks/create",
            method="POST",
            parameters={
                "task_name": "NumericEncoder",
                "fields": fields,
                "vector_name": vector_name,
            },
        )

        output = self._check_status_until_finish(
            dataset_id,
            task["task_id"],
            status_checker=status_checker,
            verbose=verbose,
            time_between_ping=time_between_ping,
        )
        return output

    def create_encode_categories_task(
        self,
        dataset_id: str,
        fields: list,
        status_checker: bool = True,
        verbose: bool = True,
        time_between_ping: int = 10,
    ):
        """
        Within a collection encode the specified array field in every document into vectors. \n
        For example, array that represents a movie's categories:
        >>> document 1 array field: {"category" : ["sci-fi", "thriller", "comedy"]}
        >>> document 2 array field: {"category" : ["sci-fi", "romance", "drama"]}
        >>> -> <Encode the arrays to vectors> ->
        >>> | sci-fi | thriller | comedy | romance | drama |
        >>> |--------|----------|--------|---------|-------|
        >>> | 1      | 1        | 1      | 0       | 0     |
        >>> | 1      | 0        | 0      | 1       | 1     |
        >>> document 1 array vector: {"movie_categories_vector_": [1, 1, 1, 0, 0]}
        >>> document 2 array vector: {"movie_categories_vector_": [1, 0, 0, 1, 1]}

        Parameters
        ----------
        dataset_id : string
            Unique name of dataset
        fields : list
            The numeric fields to encode into vectors.
        """
        task = self.make_http_request(
            endpoint=f"/datasets/{dataset_id}/tasks/create",
            method="POST",
            parameters={"task_name": "CategoriesEncoder", "fields": fields},
        )

        output = self._check_status_until_finish(
            dataset_id,
            task["task_id"],
            status_checker=status_checker,
            verbose=verbose,
            time_between_ping=time_between_ping,
        )
        return output

    def create_encode_text_task(
        self,
        dataset_id: str,
        field: str,
        alias: str = "default",
        refresh: bool = False,
        status_checker: bool = True,
        verbose: bool = True,
        time_between_ping: int = 10,
    ):
        """
        Start a task which encodes a text field
        Parameters
        ----------
        dataset_id : string
            Unique name of dataset
        field : string
            The field to encode
        alias: string
            Alias used to name a vector field. Belongs in field_{alias}vector
        refresh: bool
            Whether to rerun task on the whole dataset or just the ones missing the output
        """

        task = self.make_http_request(
            endpoint=f"/datasets/{dataset_id}/tasks/create",
            method="POST",
            parameters={
                "task_name": "TextEncoder",
                "dataset_id": dataset_id,
                "field": field,
                "alias": alias,
                "refresh": refresh,
            },
        )

        output = self._check_status_until_finish(
            dataset_id,
            task["task_id"],
            status_checker=status_checker,
            verbose=verbose,
            time_between_ping=time_between_ping,
        )
        return output

    def create_encode_textimage_task(
        self,
        dataset_id: str,
        field: str,
        alias: str = "default",
        refresh: bool = False,
        status_checker: bool = True,
        verbose: bool = True,
        time_between_ping: int = 10,
    ):
        """
        Start a task which encodes a text field for image representation
        Parameters
        ----------
        dataset_id : string
            Unique name of dataset
        field : string
            The field to encode
        alias: string
            Alias used to name a vector field. Belongs in field_{alias}vector
        refresh: bool
            Whether to rerun task on the whole dataset or just the ones missing the output
        """
        task = self.make_http_request(
            endpoint=f"/datasets/{dataset_id}/tasks/create",
            method="POST",
            parameters={
                "task_name": "TextImageEncoder",
                "dataset_id": dataset_id,
                "field": field,
                "alias": alias,
                "refresh": refresh,
            },
        )

        output = self._check_status_until_finish(
            dataset_id,
            task["task_id"],
            status_checker=status_checker,
            verbose=verbose,
            time_between_ping=time_between_ping,
        )
        return output

    def create_encode_imagetext_task(
        self,
        dataset_id: str,
        field: str,
        alias: str = "default",
        refresh: bool = False,
        status_checker: bool = True,
        verbose: bool = True,
        time_between_ping: int = 10,
    ):
        """
        Start a task which encodes an image field for text representation
        Parameters
        ----------
        dataset_id : string
            Unique name of dataset
        field : string
            The field to encode
        alias: string
            Alias used to name a vector field. Belongs in field_{alias}vector
        refresh: bool
            Whether to rerun task on the whole dataset or just the ones missing the output
        """
        task = self.make_http_request(
            endpoint=f"/datasets/{dataset_id}/tasks/create",
            method="POST",
            parameters={
                "task_name": "ImageTextEncoder",
                "dataset_id": dataset_id,
                "field": field,
                "alias": alias,
                "refresh": refresh,
            },
        )

        output = self._check_status_until_finish(
            dataset_id,
            task["task_id"],
            status_checker=status_checker,
            verbose=verbose,
            time_between_ping=time_between_ping,
        )
        return output



"""
Users can detect bias in a model by seeing which concepts certain vectors are closer to.
This is a particularly useful tool when users are looking at semantic vectors and
would like to check if certain words are leaning particularly towards any
specific category.

An example of analysing gender bias inside Google's Universal Sentence Encoder
can be found below.

.. code-block::

    # Set up the encoder
    !pip install -q vectorhub[encoders-text-tfhub]
    from vectorhub.encoders.text.tfhub import USE2Vec
    enc = USE2Vec()

    from relevanceai.bias_detection import bias_indicator

    bias_indicator(
        ["boy", "girl"], # the categories of bias
        ["basketball", "draft", "skirt", "dress", "grave digger"], # words to care about
        enc.encode
    )

"""

from typing import List, Dict, Any, Callable
from doc_utils import DocUtils
from relevanceai.utils.decorators.analytics import track_event_usage

try:
    import matplotlib.pyplot as plt
except ModuleNotFoundError:
    # Don't trigger if missing
    pass


class BiasIndicator(DocUtils):
    def displacement_from_unit_line(self, v):
        """Distance from unit line"""
        proj = v[0] / 2 ** (1 / 2), v[1] / 2 ** (1 / 2)
        # Use the TanH activation function to prevent blow-up
        return (v[0] - proj[0]) ** 2 - (v[1] - proj[1]) ** 2

    def remove_box_line(self, axes, edge: str):
        axes[1].spines["top"].set_visible(False)

    def bias_indicator_to_html(
        self,
        anchor_documents: List[Dict],
        documents: List[Dict],
        metadata_field: str,
        vector_field: str,
        marker_colors=["purple", "#2E8B86"],
        white_sep_space: float = 0.4,
        xlabel="L2 Distance From Neutrality Line",
        title: str = None,
    ):
        fig, axes = plt.subplots(ncols=2, sharey=True)
        anchor_values = self.get_field_across_documents(
            metadata_field, anchor_documents
        )
        graph_data = self.bias_indicator(
            anchor_documents=anchor_documents,
            documents=documents,
            metadata_field=metadata_field,
            vector_field=vector_field,
        )
        x1 = graph_data[
            self._get_bias_title(anchor_documents[0], metadata_field=metadata_field)
        ]
        x2 = graph_data[
            self._get_bias_title(anchor_documents[1], metadata_field=metadata_field)
        ]
        metadata = graph_data["metadata"]
        axes[0].barh(
            metadata, [abs(x) for x in x1], align="center", color=marker_colors[0]
        )
        axes[0].set(title=anchor_values[1])
        axes[1].barh(metadata, x2, align="center", color=marker_colors[1], zorder=10)
        axes[1].set(title=anchor_values[0])
        axes[0].invert_xaxis()
        axes[0].set(yticklabels=metadata)
        axes[0].yaxis.tick_right()
        axes[1].spines["right"].set_visible(False)
        axes[1].spines["top"].set_visible(False)
        axes[0].spines["left"].set_visible(False)
        axes[0].spines["top"].set_visible(False)
        axes[0].set_xlabel(xlabel)
        axes[1].set_xticks(axes[0].get_xticks())
        fig.tight_layout()
        fig.subplots_adjust(wspace=white_sep_space)
        if title is None:
            fig.suptitle(
                f'Bias Indicator Between "{anchor_values[1]}" and "{anchor_values[0]}"',
                y=1.1,
            )
        else:
            fig.suptitle(title)

    def _get_bias_title(self, document, metadata_field):
        return f"bias_towards_{self.get_field(metadata_field, document)}"

    def bias_indicator(
        self,
        anchor_documents: List[Dict],
        documents: List[Dict],
        metadata_field: str,
        vector_field: str,
    ):
        """
        Bias Indicator returns a 0 if it is not the most similar group and a displacement value
        from neutrality towards the most similar group.
        """
        metadata = self.get_field_across_documents(metadata_field, documents)
        x = self.get_cosine_similarity_scores(
            documents, anchor_documents[0], vector_field
        )
        y = self.get_cosine_similarity_scores(
            documents, anchor_documents[1], vector_field
        )
        v = list(zip(x, y))
        displacement = [self.displacement_from_unit_line(_v) for _v in v]
        x1 = [min(_d, 0) for _d in displacement]
        x2 = [max(_d, 0) for _d in displacement]
        response = {}
        title = self._get_bias_title(anchor_documents[0], metadata_field)
        response[title] = x1
        title = self._get_bias_title(anchor_documents[1], metadata_field)
        response[title] = x2
        response["metadata"] = metadata
        return response

    def get_cosine_similarity_scores(
        self,
        documents: List[Dict[str, Any]],
        anchor_document: Dict[str, Any],
        vector_field: str,
    ) -> List[float]:
        """
        Compare scores based on cosine similarity

        Args:
            other_documents:
                List of documents (Python Dictionaries)
            anchor_document:
                Document to compare all the other documents with.
            vector_field:
                The field in the documents to compare
        Example:
            >>> documents = [{...}]
            >>> ViClient.get_cosine_similarity_scores(documents[1:10], documents[0])
        """
        similarity_scores = []
        for i, doc in enumerate(documents):
            similarity_score = self.calculate_cosine_similarity(
                self.get_field(vector_field, doc),
                self.get_field(vector_field, anchor_document),
            )
            similarity_scores.append(similarity_score)
        return similarity_scores

    @staticmethod
    def calculate_cosine_similarity(a, b):
        from numpy import inner
        from numpy.linalg import norm
        from scipy import spatial

        return 1 - spatial.distance.cosine(a, b)
        # return inner(a, b) / (norm(a) * norm(b))


@track_event_usage("bias_indicator")
def bias_indicator(anchors: List, values: List, model: Callable):
    """
    Simple bias indicator based on vectors.

    .. code-block::

        from relevanceai.bias_detection import bias_indicator
        from vectorhub.encoders.text.tfhub import USE2Vec
        enc = USE2Vec()

        bias_indicator(
            ["boy", "girl"],
            ["basketball", "draft", "skirt", "dress", "grave digger"],
            enc.encode
        )

    """
    # create the relevant documents
    anchor_docs = [{"value": a, "_vector_": model(a)} for a in anchors]
    value_docs = [{"value": a, "_vector_": model(a)} for a in values]
    return BiasIndicator().bias_indicator_to_html(
        anchor_documents=anchor_docs,
        documents=value_docs,
        metadata_field="value",
        vector_field="_vector_",
    )



"""API Client
"""
from relevanceai.client.helpers import Credentials
from relevanceai.utils.base import _Base

from relevanceai._api.endpoints.admin.admin import AdminClient
from relevanceai._api.endpoints.datasets.datasets import DatasetsClient
from relevanceai._api.endpoints.services.services import ServicesClient
from relevanceai._api.endpoints.reports.reports import ReportsClient
from relevanceai._api.endpoints.deployables.deployables import DeployableClient

from relevanceai.constants.errors import FieldNotFoundError

from relevanceai.utils.datasets import ExampleDatasets
from relevanceai.utils import make_id

from doc_utils import DocUtils


class APIEndpointsClient(_Base, DocUtils):
    """API Client"""

    def __init__(
        self,
        credentials: Credentials,
        **kwargs,
    ):
        self.datasets = DatasetsClient(credentials)
        self.services = ServicesClient(credentials)
        self.example_datasets = ExampleDatasets()
        self.admin = AdminClient(credentials)
        self.reports = ReportsClient(credentials)
        self.deployables = DeployableClient(credentials)
        super().__init__(
            credentials=credentials,
            **kwargs,
        )

    def _convert_id_to_string(self, documents, create_id: bool = False):
        try:
            self.set_field_across_documents(
                "_id", [str(i["_id"]) for i in documents], documents
            )
        except KeyError:
            if create_id:
                self.set_field_across_documents(
                    "_id", [make_id(document) for document in documents], documents
                )
            else:
                raise FieldNotFoundError(
                    "Missing _id field. Set `create_id=True` to automatically generate IDs."
                )

    def _are_fields_in_schema(self, fields, dataset_id, schema=None):
        """
        Check fields are in schema
        """
        if schema is None:
            schema = self.datasets.schema(dataset_id)
        invalid_fields = []
        for i in fields:
            if i not in schema:
                invalid_fields.append(i)
        if len(invalid_fields) > 0:
            raise ValueError(
                f"{', '.join(invalid_fields)} are invalid fields. They are not in the dataset schema."
            )
        return



# -*- coding: utf-8 -*-
"""
Pandas like dataset API
"""
import warnings
from relevanceai.constants.warning import Warning
import itertools

from itertools import chain
from collections import Counter

from typing import Callable, Dict, List, Optional

from tqdm.auto import tqdm

from relevanceai.dataset.write.write import Write

from relevanceai.operations.vector.local_nearest_neighbours import NearestNeighbours
from relevanceai.operations.preprocessing.text.base_text_processing import MLStripper
from relevanceai.operations.cluster.constants import NEAREST_NEIGHBOURS

from relevanceai.utils.decorators.analytics import track
from relevanceai.utils.logger import FileLogger
from relevanceai.utils.decorators.version import beta

# TODO: Separate out operations into different files - cluster/search/dr


class Labels(Write):
    @track
    def label_vector(
        self,
        vector,
        alias: str,
        label_dataset_id: str,
        label_vector_field: str,
        label_fields: list,
        number_of_labels: int = 1,
        similarity_metric: NEAREST_NEIGHBOURS = "cosine",
        score_field: str = "_search_score",
        **kwargs,
    ):
        """
        Label a dataset based on a model.

        .. warning::
            This function is currently in beta and is likely to change in the future.
            We recommend not using this in any production systems.

        .. note::
            **New in v0.32.0**

        Parameters
        -------------

        vector_fields: list
            The list of vector field
        label_dataset_id: str
            The dataset to label with
        alias: str
            The alias of the labels (for example - "ranking_labels")
        label_dataset_id: str
            The dataset to use fo rlabelling
        label_vector_field: str
            The vector field of the label dataset
        label_fields: list
            The label field of the dataset to use
        number_of_labels: int
            The numebr of labels to get
        similarity_metric: str
            The similarity metric to adopt
        score_field: str
            The field to use for scoring

        Example
        -------------

        .. code-block::


            from relevanceai import Client
            from relevanceai.ops.clusterops.cluster import ClusterOps
            from relevanceai.ops.clusterops.kmeans_clusterer import KMeansModel

            client = Client()

            dataset_id = "sample_dataset_id"
            df = client.Dataset(dataset_id)

            result = df.label_vector(
                [...],
                label_vector_field="sample_1_vector_",
                alias="alias_sample",
                label_dataset_id=label_dataset_id_id,
                label_fields=["sample_1_label"],
                number_of_labels=1,
            )

        """
        # Download documents in the label dataset
        label_documents: list = self._get_all_documents(
            label_dataset_id, select_fields=[label_vector_field] + label_fields
        )

        # Build a index
        labels = self._get_nearest_labels(
            label_documents=label_documents,
            vector=vector,
            label_vector_field=label_vector_field,
            similarity_metric=similarity_metric,
            number_of_labels=number_of_labels,
            score_field=score_field,
            label_fields=label_fields,
        )

        # Store things according to
        # {"_label_": {"field": {"alias": [{"label": 3, "similarity_score": 0.4}]}
        return self.store_labels_in_document(labels, alias)

    @track
    def store_labels_in_document(self, labels: list, alias: str):
        if isinstance(labels, dict) and "label" in labels:
            return {"_label_": {alias: labels["label"]}}
        return {"_label_": {alias: labels}}

    def _get_nearest_labels(
        self,
        label_documents: List[Dict],
        vector: List[float],
        label_vector_field: str,
        similarity_metric: NEAREST_NEIGHBOURS,
        number_of_labels: int,
        label_fields: List[str],
        score_field="_label_score",
    ):
        nearest_neighbors: List[Dict] = NearestNeighbours.get_nearest_neighbours(
            label_documents,
            vector,
            label_vector_field,
            similarity_metric,
            score_field=score_field,
        )[:number_of_labels]
        labels: List[Dict] = self.subset_documents(
            [score_field] + label_fields, nearest_neighbors
        )
        # Preprocess labels for easier frontend access
        new_labels = {}
        for lf in label_fields:
            new_labels[lf] = [
                {"label": l.get(lf), score_field: l.get(score_field)} for l in labels
            ]
        return new_labels

    @track
    def label_document(
        self,
        document: dict,
        vector_field: str,
        vector: List[float],
        alias: str,
        label_dataset_id: str,
        label_vector_field: str,
        label_fields: List[str],
        number_of_labels: int = 1,
        similarity_metric="cosine",
        score_field: str = "_search_score",
    ):
        """
        Label a dataset based on a model.

        .. warning::
            This function is currently in beta and is likely to change in the future.
            We recommend not using this in any production systems.

        .. note::
            **New in v0.32.0**

        Parameters
        -------------

        document: dict
            A document to label
        vector_field: str
            The list of vector field
        label_dataset_id: str
            The dataset to label with
        alias: str
            The alias of the labels (for example - "ranking_labels")
        label_dataset_id: str
            The dataset to use fo rlabelling
        label_vector_field: str
            The vector field of the label dataset
        label_fields: list
            The label field of the dataset to use
        number_of_labels: int
            The numebr of labels to get
        similarity_metric: str
            The similarity metric to adopt
        score_field: str
            The field to use for scoring

        Example
        ---------

        .. code-block::

            from relevanceai import Client
            client = Client()
            df = client.Dataset("sample_dataset_id")

            results = df.label_document(
                document={...},
                vector_field="sample_1_vector_",
                alias="example",
                label_dataset_id=label_dataset_id_id,
                label_fields=["sample_1_label"],
                label_vector_field="sample_1_vector_",
                filters=[
                    {
                        "field": "sample_1_label",
                        "filter_type": "exists",
                        "condition": ">=",
                        "condition_value": " ",
                    },
                ],
            )
        """
        vector = self.get_field(vector_field, document)
        labels = self.label_vector(
            vector_field=vector_field,
            vector=vector,
            alias=alias,
            label_dataset_id=label_dataset_id,
            label_vector_field=label_vector_field,
            label_fields=label_fields,
            number_of_labels=number_of_labels,
            score_field=score_field,
            similarity_metric=similarity_metric,
        )
        document.update(self.store_labels_in_document(labels, alias))
        return document

    @track
    def label_from_dataset(
        self,
        vector_field: str,
        alias: str,
        label_dataset_id: str,
        label_vector_field: str,
        label_fields: List[str],
        number_of_labels: int = 1,
        filters: Optional[list] = None,
        similarity_metric="cosine",
        score_field: str = "_search_score",
    ):
        """
        Label a dataset based on a model.

        .. warning::
            This function is currently in beta and is likely to change in the future.
            We recommend not using this in any production systems.

        .. note::
            **New in v0.32.0**

        Parameters
        -------------

        vector_field: str
            The vector field to match with
        alias: str
            The alias of the labels (for example - "ranking_labels")
        label_dataset_id: str
            The dataset to use fo rlabelling
        label_vector_field: str
            The vector field of the label dataset
        label_fields: list
            The label field of the dataset to use
        filters: list
            The filters to apply to label
        number_of_labels: int
            The numebr of labels to get
        similarity_metric: str
            The similarity metric to adopt
        score_field: str
            The field to use for scoring

        Example
        ----------

        .. code-block::

            from relevanceai import Client
            client = Client()
            df = client.Dataset("sample_dataset_id")

            results = df.label(
                vector_field="sample_1_vector_",
                alias="example",
                label_dataset_id=label_dataset_id,
                label_fields=["sample_1_label"],
                label_vector_field="sample_1_vector_",
                filters=[
                    {
                        "field": "sample_1_label",
                        "filter_type": "exists",
                        "condition": ">=",
                        "condition_value": " ",
                    },
                ],
            )

        """
        filters = [] if filters is None else filters

        # Download documents in the label dataset
        filters += [
            {
                "field": label_field,
                "filter_type": "exists",
                "condition": ">=",
                "condition_value": " ",
            }
            for label_field in label_fields
        ] + [
            {
                "field": label_vector_field,
                "filter_type": "exists",
                "condition": ">=",
                "condition_value": " ",
            },
        ]
        label_documents: list = self._get_all_documents(
            label_dataset_id,
            select_fields=[label_vector_field] + label_fields,
            filters=filters,
        )

        def label_and_store(d: dict):
            labels = self._get_nearest_labels(
                label_documents=label_documents,
                vector=self.get_field(vector_field, d),
                label_vector_field=label_vector_field,
                similarity_metric=similarity_metric,
                number_of_labels=number_of_labels,
                score_field=score_field,
                label_fields=label_fields,
            )
            d.update(self.store_labels_in_document(labels, alias))
            return d

        def bulk_label_documents(documents):
            [label_and_store(d) for d in documents]
            return documents

        return self.bulk_apply(
            bulk_label_documents,
            filters=[
                {
                    "field": vector_field,
                    "filter_type": "exists",
                    "condition": ">=",
                    "condition_value": " ",
                },
            ],
        )

    @track
    def label_from_list(
        self,
        vector_field: str,
        model: Callable,
        label_list: list,
        similarity_metric="cosine",
        number_of_labels: int = 1,
        score_field: str = "_search_score",
        alias: Optional[str] = None,
    ):
        """Label from a given list.

        Parameters
        ------------

        vector_field: str
            The vector field to label in the original dataset
        model: Callable
            This will take a list of strings and then encode them
        label_list: List
            A list of labels to accept
        similarity_metric: str
            The similarity metric to accept
        number_of_labels: int
            The number of labels to accept
        score_field: str
            What to call the scoring of the labels
        alias: str
            The alias of the labels

        Example
        --------

        .. code-block::

            from relevanceai import Client
            client = Client()
            df = client.Dataset("sample")

            # Get a model to help us encode
            from vectorhub.encoders.text.tfhub import USE2Vec
            enc = USE2Vec()

            # Use that model to help with encoding
            label_list = ["dog", "cat"]

            df = client.Dataset("_github_repo_vectorai")

            df.label_from_list("documentation_vector_", enc.bulk_encode, label_list, alias="pets")

        """
        if alias is None:
            warnings.warn(
                "No alias is detected for labelling. Default to 'default' as the alias."
            )
            alias = "default"
        print("Encoding labels...")
        label_vectors = []
        for c in self.chunk(label_list, chunksize=20):
            with FileLogger(verbose=True):
                label_vectors.extend(model(c))

        if len(label_vectors) == 0:
            raise ValueError("Failed to encode.")

        # we need this to mock label documents - these values are not important
        # and can be changed :)
        LABEL_VECTOR_FIELD = "label_vector_"
        LABEL_FIELD = "label"

        label_documents = [
            {LABEL_VECTOR_FIELD: label_vectors[i], LABEL_FIELD: label}
            for i, label in enumerate(label_list)
        ]

        return self._bulk_label_dataset(
            label_documents=label_documents,
            vector_field=vector_field,
            label_vector_field=LABEL_VECTOR_FIELD,
            similarity_metric=similarity_metric,
            number_of_labels=number_of_labels,
            score_field=score_field,
            label_fields=[LABEL_FIELD],
            alias=alias,
        )

    def _bulk_label_dataset(
        self,
        label_documents,
        vector_field,
        label_vector_field,
        similarity_metric,
        number_of_labels,
        score_field,
        label_fields,
        alias,
    ):
        def label_and_store(d: dict):
            labels = self._get_nearest_labels(
                label_documents=label_documents,
                vector=self.get_field(vector_field, d),
                label_vector_field=label_vector_field,
                similarity_metric=similarity_metric,
                number_of_labels=number_of_labels,
                score_field=score_field,
                label_fields=label_fields,
            )
            d.update(self.store_labels_in_document(labels, alias))
            return d

        def bulk_label_documents(documents):
            [label_and_store(d) for d in documents]
            return documents

        print("Labelling dataset...")
        return self.bulk_apply(
            bulk_label_documents,
            filters=[
                {
                    "field": vector_field,
                    "filter_type": "exists",
                    "condition": ">=",
                    "condition_value": " ",
                },
            ],
            select_fields=[vector_field],
        )

    def _set_up_nltk(
        self,
        stopwords_dict: str = "english",
        additional_stopwords: Optional[list] = None,
    ):
        """Additional stopwords to include"""
        import nltk
        from nltk.corpus import stopwords

        additional_stopwords = (
            [] if additional_stopwords is None else additional_stopwords
        )

        self._is_set_up = True
        nltk.download("stopwords")
        nltk.download("punkt")
        self.eng_stopwords = stopwords.words(stopwords_dict)
        self.eng_stopwords.extend(additional_stopwords)
        self.eng_stopwords = set(self.eng_stopwords)

    def clean_html(self, html):
        """Cleans HTML from text"""
        s = MLStripper()
        if html is None:
            return ""
        s.feed(html)
        return s.get_data()

    def get_word_count(self, text_fields: List[str]):
        """
        Create labels from a given text field.

        Parameters
        ------------

        text_fields: list
            List of text fields


        Example
        ------------

        .. code-block::

            from relevanceai import Client
            client = Client()
            df = client.Dataset("sample")
            df.get_word_count()

        """
        import nltk
        from nltk.corpus import stopwords

        raise NotImplementedError

    def generate_text_list_from_documents(
        self,
        documents: Optional[list] = None,
        text_fields: Optional[list] = None,
        clean_html: bool = False,
    ):
        """
        Generate a list of text from documents to feed into the counter
        model.
        Parameters
        -------------
        documents: list
            A list of documents
        fields: list
            A list of fields
        clean_html: bool
            If True, also cleans the text in a given text document to remove HTML. Will be slower
            if processing on a large document
        """
        documents = [] if documents is None else documents
        text_fields = [] if text_fields is None else text_fields

        text = self.get_fields_across_documents(
            text_fields, documents, missing_treatment="ignore"
        )
        return list(itertools.chain.from_iterable(text))

    def generate_text_list(
        self,
        filters: Optional[list] = None,
        batch_size: int = 20,
        text_fields: Optional[list] = None,
        cursor: str = None,
    ):
        filters = [] if filters is None else filters
        text_fields = [] if text_fields is None else text_fields

        filters += [
            {
                "field": tf,
                "filter_type": "exists",
                "condition": "==",
                "condition_value": " ",
            }
            for tf in text_fields
        ]
        documents = self.get_documents(
            batch_size=batch_size,
            select_fields=text_fields,
            filters=filters,
            cursor=cursor,
        )
        return self.generate_text_list_from_documents(
            documents, text_fields=text_fields
        )

    def get_ngrams(
        self,
        text,
        n: int = 2,
        stopwords_dict: str = "english",
        additional_stopwords: Optional[list] = None,
        min_word_length: int = 2,
        preprocess_hooks: Optional[list] = None,
    ):
        additional_stopwords = (
            [] if additional_stopwords is None else additional_stopwords
        )
        preprocess_hooks = [] if preprocess_hooks is None else preprocess_hooks

        try:
            return self._get_ngrams(
                text=text,
                n=n,
                additional_stopwords=additional_stopwords,
                min_word_length=min_word_length,
                preprocess_hooks=preprocess_hooks,
            )
        except:
            # Specify that this shouldn't necessarily error out.
            self.set_up(
                stopwords_dict=stopwords_dict, additional_stopwords=additional_stopwords
            )
            return self._get_ngrams(
                text=text,
                n=n,
                min_word_length=min_word_length,
                preprocess_hooks=preprocess_hooks,
            )

    def _get_ngrams(
        self,
        text,
        n: int = 2,
        additional_stopwords: Optional[list] = None,
        min_word_length: int = 2,
        preprocess_hooks: Optional[list] = None,
    ):
        """Get the bigrams"""
        from nltk import word_tokenize
        from nltk.util import ngrams

        additional_stopwords = (
            [] if additional_stopwords is None else additional_stopwords
        )
        preprocess_hooks = [] if preprocess_hooks is None else preprocess_hooks

        if additional_stopwords:
            [self.eng_stopwords.add(s) for s in additional_stopwords]

        n_grams = []
        for line in text:
            for p_hook in preprocess_hooks:
                line = p_hook(line)
            token = word_tokenize(line)
            n_grams.append(list(ngrams(token, n)))

        def length_longer_than_min_word_length(x):
            return len(x.strip()) >= min_word_length

        def is_not_stopword(x):
            return x.strip() not in self.eng_stopwords

        def is_clean(text_list):
            return all(
                length_longer_than_min_word_length(x) and is_not_stopword(x)
                for x in text_list
            )

        counter = Counter([" ".join(x) for x in chain(*n_grams) if is_clean(x)])
        return counter
        # return dict(counter.most_common(most_common))

    @track
    @beta
    def keyphrases(
        self,
        text_fields: list,
        algorithm: str = "rake",
        n: int = 2,
        most_common: int = 10,
        filters: Optional[list] = None,
        additional_stopwords: Optional[list] = None,
        min_word_length: int = 2,
        batch_size: int = 1000,
        document_limit: int = None,
        preprocess_hooks: Optional[List[callable]] = None,
        verbose: bool = True,
    ) -> list:
        """
        Returns the most common phrase in the following format:

        .. code-block::

            [('heavily draping faux fur', 16.0),
            ('printed sweatshirt fabric made', 14.333333333333334),
            ('high paper bag waist', 14.25),
            ('ribbed organic cotton jersey', 13.803030303030303),
            ('soft sweatshirt fabric', 9.0),
            ('open back pocket', 8.5),
            ('layered tulle skirt', 8.166666666666666),
            ('soft brushed inside', 8.0),
            ('discreet side pockets', 7.5),
            ('cotton blend', 5.363636363636363)]

        Parameters
        ------------

        text_fields: list
            A list of text fields
        algorithm: str
            The algorithm to use. Must be one of `nltk` or `rake`.
        n: int
            if algorithm is `nltk`, this will set the number of words. If `rake`, then it
            will do nothing.
        most_common: int
            How many to return
        filters: list
            A list of filters to supply
        additional_stopwords: list
            A list of additional stopwords to supply
        min_word_length: int
            The minimum word length to apply to clean. This can be helpful if there are common
            acronyms that you want to exclude.
        batch_size: int
            Batch size is the number of documents to retrieve in a chunk
        document_limit: int
            The maximum number of documents in a dataset
        preprocess_hooks: List[Callable]
            A list of process hooks to clean text before they count as a word

        Example
        ----------

        .. code-block::

            from relevanceai import Client
            client = Client()
            ds = client.Dataset("sample")
            # Returns the top keywords in a text field
            ds.keyphrases(text_fields=["sample"])


            # Create an e-commerce dataset

            from relevanceai.package_utils.datasets import get_dummy_ecommerce_dataset
            docs = get_dummy_ecommerce_dataset()
            ds = client.Dataset("ecommerce-example")
            ds.upsert_documents(docs)
            ds.keyphrases(text_fields=text_fields, algorithm="nltk", n=3)
            def remove_apostrophe(string):
                return string.replace("'s", "")
            ds.keyphrases(text_fields=text_fields, algorithm="nltk", n=3, preprocess_hooks=[remove_apostrophe])
            ds.keyphrases(text_fields=text_fields, algorithm="nltk", n=3, additional_stopwords=["Men", "Women"])

        """
        self._check_keyphrase_algorithm_requirements(algorithm)
        filters = [] if filters is None else filters
        additional_stopwords = (
            [] if additional_stopwords is None else additional_stopwords
        )
        preprocess_hooks = [] if preprocess_hooks is None else preprocess_hooks

        counter: Counter = Counter()
        if not hasattr(self, "_is_set_up"):
            if verbose:
                print("setting up NLTK...")
            self._set_up_nltk()

        # Mock a dummy documents so I can loop immediately without weird logic
        documents: dict = {"documents": [[]], "cursor": None}
        if verbose:
            print("Updating word count...")
        while len(documents["documents"]) > 0 and (
            document_limit is None or sum(counter.values()) < document_limit
        ):
            # TODO: make this into a progress bar instead
            documents = self.get_documents(
                filters=filters,
                cursor=documents["cursor"],
                batch_size=batch_size,
                select_fields=text_fields,
                include_cursor=True,
            )
            string = self.generate_text_list_from_documents(
                documents=documents["documents"],
                text_fields=text_fields,
            )

            if algorithm == "nltk":
                ngram_counter = self._get_ngrams(
                    string,
                    n=n,
                    additional_stopwords=additional_stopwords,
                    min_word_length=min_word_length,
                    preprocess_hooks=preprocess_hooks,
                )
            elif algorithm == "rake":
                ngram_counter = self._get_rake_keyphrases(
                    string,
                )
            counter.update(ngram_counter)
        return counter.most_common(most_common)

    def _check_keyphrase_algorithm_requirements(self, algorithm: str):
        if algorithm == "rake":
            try:
                import rake_nltk
            except ModuleNotFoundError:
                raise ModuleNotFoundError("Run `pip install rake-nltk`.")
        elif algorithm == "nltk":
            try:
                import nltk
            except ModuleNotFoundError:
                raise ModuleNotFoundError("Run `pip install nltk`")

    def _get_rake_keyphrases(self, string, **kw):
        from rake_nltk import Rake

        r = Rake(**kw)
        r.extract_keywords_from_sentences(string)
        results = r.get_ranked_phrases_with_scores()
        return {v: k for k, v in results}

    def cluster_keyphrases(
        self,
        vector_fields: List[str],
        text_fields: List[str],
        cluster_alias: str,
        cluster_field: str = "_cluster_",
        num_clusters: int = 100,
        most_common: int = 10,
        preprocess_hooks: Optional[List[callable]] = None,
        algorithm: str = "rake",
        n: int = 2,
    ):
        """
        Simple implementation of the cluster word cloud.

        Parameters
        ------------
        vector_fields: list
            The list of vector fields
        text_fields: list
            The list of text fields
        cluster_alias: str
            The alias of the cluster
        cluster_field: str
            The cluster field to try things on
        num_clusters: int
            The number of clusters
        preprocess_hooks: list
            The preprocess hooks
        algorithm: str
            The algorithm to use
        n: int
            The number of words

        """
        preprocess_hooks = [] if preprocess_hooks is None else preprocess_hooks

        vector_fields_str = ".".join(sorted(vector_fields))
        field = f"{cluster_field}.{vector_fields_str}.{cluster_alias}"
        all_clusters = self.facets([field], page_size=num_clusters)
        cluster_counters = {}
        if "results" in all_clusters:
            all_clusters = all_clusters["results"]
        # TODO: Switch to multiprocessing
        for c in tqdm(all_clusters[field]):
            cluster_value = c[field]
            top_words = self.keyphrases(
                text_fields=text_fields,
                n=n,
                filters=[
                    {
                        "field": field,
                        "filter_type": "contains",
                        "condition": "==",
                        "condition_value": cluster_value,
                    }
                ],
                most_common=most_common,
                preprocess_hooks=preprocess_hooks,
                algorithm=algorithm,
            )
            cluster_counters[cluster_value] = top_words
        return cluster_counters

    # TODO: Add keyphrases to auto cluster
    # def auto_cluster_keyphrases(
    #     vector_fields: List[str],
    #     text_fields: List[str],
    #     cluster_alias: str,
    #     deployable_id: str,
    #     n: int = 2,
    #     cluster_field: str = "_cluster_",
    #     num_clusters: int = 100,
    #     preprocess_hooks: Optional[List[callable]] = None,
    # ):
    #     """
    #     # TODO:
    #     """
    #     pass

    def _add_cluster_word_cloud_to_config(self, data, cluster_value, top_words):
        # TODO: Add this to wordcloud deployable
        # hacky way I implemented to add top words to config
        data["configuration"]["cluster-labels"][cluster_value] = ", ".join(
            [k for k in top_words if k != "machine learning"]
        )
        data["configuration"]["cluster-descriptions"][cluster_value] = str(top_words)

    @track
    def label_from_common_words(
        self,
        text_field: str,
        model: Callable = None,
        most_common: int = 1000,
        n_gram: int = 1,
        temp_vector_field: str = "_label_vector_",
        labels_fn="labels.txt",
        stopwords: Optional[list] = None,
        algorithm: str = "nltk",
    ):
        """
        Label by the most popular keywords.

        Algorithm:

        - Get top X keywords or bigram for a text field
        - Default X to 1000 or something scaled towards number of documents
        - Vectorize those into keywords
        - Label every document with those top keywords

        .. note::
            **New in v1.1.0**

        Parameters
        ------------

        text_fields: str
            The field to label
        model: Callable
            The function or callable to turn text into a vector.
        most_common: int
            How many of the most common worsd do you want to use as labels
        n_gram: int
            How many word co-occurrences do you want to consider
        temp_vector_field: str
            The temporary vector field name
        labels_fn: str
            The filename for labels to be saved in.
        stopwords: list
            A list of stopwords
        algorithm: str
            The algorithm to use. Must be one of `nltk` or `rake`.

        Example
        --------

        .. code-block::

            import random
            from relevanceai import Client
            from relevanceai.package_utils.datasets import mock_documents
            from relevanceai.package_utils.logger import FileLogger

            client = Client()
            ds = client.Dataset("sample")
            documents = mock_documents()
            ds.insert_documents(documents)

            def encode():
                return [random.randint(0, 100) for _ in range(5)]

            ds.label_from_common_words(
                text_field="sample_1_label",
                model=encode,
                most_common=10,
                n_gram=1
            )

        """
        stopwords = [] if stopwords is None else stopwords

        labels = self.keyphrases(
            text_fields=[text_field],
            n=n_gram,
            most_common=most_common,
            additional_stopwords=stopwords,
            algorithm=algorithm,
        )

        with open(labels_fn, "w") as f:
            f.write(str(labels))

        print(f"Saved labels to {labels_fn}")
        # Add support if already encoded
        def encode(doc):
            with FileLogger():
                doc[temp_vector_field] = model(self.get_field(text_field, doc))
            return doc

        self.apply(encode, select_fields=[text_field])
        # create_labels_from_text_field
        return self.label_from_list(
            vector_field=temp_vector_field,
            model=model,
            label_list=[x[0] for x in labels],
        )



import asyncio
import time
import traceback
import warnings

import orjson

from datetime import datetime
from concurrent.futures import as_completed, wait
from pathlib import Path
from threading import Thread
from typing import Callable, Optional, Tuple

from appdirs import user_cache_dir
from relevanceai._api.batch.retrieve import BatchRetrieveClient
from relevanceai._api.endpoints.api_client import APIEndpointsClient
from relevanceai.utils.logger import FileLogger
from relevanceai.utils.progress_bar import progress_bar
from relevanceai.constants.warning import Warning

MB_TO_BYTE = 1024 * 1024


class EventLoop(Thread):
    """
    The event loop is required to call asynchronous functions from
    a regular (synchronous) function. This class spins off a
    background thread that runs an event loop. All asynchronous
    tasks are run there.
    """

    def __init__(self, tasks, show_progress_bar):
        super().__init__()
        self._loop = asyncio.new_event_loop()
        self.daemon = True

        self.futures = []
        self.tasks = tasks

        self.show_progress_bar = show_progress_bar
        self.progress_bar = progress_bar(
            range(len(tasks)), show_progress_bar=show_progress_bar
        )

    def run(self):
        self._loop.run_forever()

    async def _create_future(self, task):
        return await asyncio.create_task(task)

    def execute_tasks(self):
        self.futures.extend(
            [
                asyncio.run_coroutine_threadsafe(self._create_future(task), self._loop)
                for task in self.tasks
            ]
        )

        if self.show_progress_bar:
            with self.progress_bar as progress_bar:
                for _ in as_completed(self.futures):
                    progress_bar.update(1)
        else:
            wait(self.futures)

    def terminate(self):
        self._loop.call_soon_threadsafe(self._loop.stop)
        self.join()


class BatchInsertAsyncHelpers(BatchRetrieveClient, APIEndpointsClient):
    async def _apply_bulk_fn(self, bulk_fn, documents: list):
        """
        Called from _process_documents. Calls bulk_fn on documents.

        Parameters
        ----------
        bulk_fn
            An asynchronous function that takes in the documents

        documents: list
            A list of documents
        """
        if len(documents) == 0:
            warnings.warn(Warning.NO_DOCUMENT_DETECTED)
            return {
                "inserted": 0,
                "failed_documents": [],
                "failed_documents_detailed": [],
            }

        num_documents_inserted: int = 0
        # Maintain a reference to documents to keep track during looping
        documents_remaining = documents.copy()
        for _ in range(int(self.config.get_option("retries.number_of_retries"))):
            if len(documents_remaining) > 0:
                # bulk_update_async
                response = await bulk_fn(documents=documents)
                num_documents_inserted += response["inserted"]
                documents_remaining = [
                    document
                    for document in documents_remaining
                    if document["_id"] in response["failed_documents"]
                ]
            else:
                # Once documents_remaining is empty, leave the for-loop...
                break
            # ...else, wait some amount of time before retrying
            time.sleep(int(self.config["retries.seconds_between_retries"]))

        return {
            "inserted": num_documents_inserted,
            "failed_documents": documents_remaining,
        }

    async def _process_documents(
        self,
        dataset_id: str,
        documents: list,
        insert: bool,
        use_json_encoder: bool = True,
        create_id: bool = False,
        verbose: bool = True,
        **kwargs,
    ):
        """
        Called from pull_update_push_async. This method determines via user
        input whether to insert or to update documents in a Dataset. (A
        Dataset that does not exist is automatically created.) Then, the
        operation (either bulk_insert_async or bulk_update_async) is wrapped
        by _apply_bulk_fn and then applied to the given documents.

        Parameters
        ----------
        dataset_id: str
            The dataset_id of the collection to change

        documents: list
            A list of documents

        insert: bool
            If True, inserts rather than updates an already-existing dataset

        use_json_encoder: bool
            Whether to automatically convert documents to json encodable format

        create_id: bool
            If True, creates a indices for the documents

        verbose: bool
            If True, print user-informing statements.

        **kwargs
            Additional arguments for bulk_insert_async or bulk_update_async
        """
        if use_json_encoder:
            documents = self.json_encoder(documents)

        in_dataset = dataset_id in self.datasets.list()["datasets"]
        if not in_dataset or insert:
            operation = f"inserting into {dataset_id}"
            if not in_dataset:
                self.datasets.create(dataset_id)
            if insert:
                create_id = True
            self._convert_id_to_string(documents, create_id=create_id)

            async def bulk_fn(documents):
                return await self.datasets.bulk_insert_async(
                    dataset_id=dataset_id,
                    documents=documents,
                    **{
                        key: value
                        for key, value in kwargs.items()
                        if key not in {"dataset_id", "updates"}
                    },
                )

        else:
            operation = f"updating {dataset_id}"
            self._convert_id_to_string(documents, create_id=create_id)

            async def bulk_fn(documents):
                return await self.datasets.documents.bulk_update_async(
                    dataset_id=dataset_id,
                    updates=documents,
                    **{
                        key: value
                        for key, value in kwargs.items()
                        if key not in {"dataset_id", "documents"}
                    },
                )

        self.logger.info(f"You are currently {operation}")

        self.logger.info(
            "You can track your stats and progress via our dashboard at "
            + f"https://cloud.relevance.ai/collections/dashboard/stats/?collection={dataset_id}"
        )

        if verbose:
            print(
                f"While {operation}, you can visit your dashboard at "
                + f"https://cloud.relevance.ai/dataset/{dataset_id}/dashboard/monitor/"
            )

        return await self._apply_bulk_fn(bulk_fn=bulk_fn, documents=documents)

    def _get_avg_document_size(
        self,
        dataset_id: str,
        filters: list,
        select_fields: list,
        num_documents: int = 5,
    ) -> int:
        """
        Calculates the average document size, averaged over num_documents.
        """
        documents = self.datasets.documents.get_where(
            dataset_id,
            filters=filters,
            select_fields=select_fields,
            page_size=num_documents,
        )["documents"]
        return (
            sum(map(lambda document: len(orjson.dumps(document)), documents))
            // num_documents
        )

    def _determine_optimal_chunk_size(
        self, dataset_id: str, filters: list, select_fields: list
    ) -> int:
        """
        Determines the "optimal" chunk size in terms of bytes. The optimal
        chunk size is calculated by divinding 50 MB by the average document
        size. 50 MB is a chosen constant, which is a nice size for the backend
        to handle.
        """
        document_size = self._get_avg_document_size(dataset_id, filters, select_fields)
        return (50 * MB_TO_BYTE) // document_size

    def _set_chunk_size(
        self,
        dataset_id: str,
        filters: list,
        select_fields: list,
        retrieve_chunk_size: Optional[int],
    ) -> int:
        """
        Sets the chunk size to an "optimal" size is the chunk size is not
        specified. Else, makes sure thathe chunk size is within tolerance.
        """
        if retrieve_chunk_size is None:
            chunk_size = self._determine_optimal_chunk_size(
                dataset_id, filters, select_fields
            )
            print(f"Chunk size set to {chunk_size}")
            self.logger.info(f"Chunk size set to {chunk_size}")
            retrieve_chunk_size = chunk_size

        if retrieve_chunk_size > 10_000:
            # The backend API has an upper bound of 10,000 on the number of
            # documents that are able to be pulled at once.
            chunk_size = 10_000
            self.logger.info(f"Chunk size set to {chunk_size}")
            return chunk_size
        else:
            return retrieve_chunk_size

    def _get_cache_path(self, dataset_id, select_fields) -> Path:
        """
        Gets (and creates) the cache path. The path is denoted by the hash of
        the fields as a simple means to distinguish different cache recalls
        of the same dataset.
        """
        cache_path = (
            Path(user_cache_dir())
            / Path("relevanceai")
            / Path(f"{dataset_id}-{hash(frozenset(select_fields))}")
        )
        if not cache_path.exists():
            cache_path.mkdir(parents=True)
            self.logger.info(f"Created directory {cache_path}")

        return cache_path

    def _create_subset_function(
        self,
        dataset_id: str,
        update_function: Callable[..., list],
        updating_args: dict,
        updated_dataset_id: Optional[str],
        filters: Optional[list],
        select_fields: Optional[list],
        use_json_encoder: bool,
        include_vector: bool,
        insert: bool,
        use_cache: bool,
    ):
        """
        A helper function to create the task coroutine.
        """
        cache_path = self._get_cache_path(dataset_id, select_fields)

        async def pull_update_push_subset(
            num: int, page_size: int, cursor: Optional[str]
        ):
            save_file = cache_path / f"pull-{num}.json"
            if save_file.exists() and use_cache:
                with open(save_file, "rb") as infile:
                    documents = orjson.loads(infile.read())
            else:
                response = await self.datasets.documents.get_where_async(
                    dataset_id,
                    filters=filters,
                    cursor=cursor,
                    page_size=page_size,
                    select_fields=select_fields,
                    include_vector=include_vector,
                )

                documents = response["documents"]

                with open(save_file, "wb") as outfile:
                    outfile.write(orjson.dumps(documents))

            try:
                updated_documents = update_function(documents, **updating_args)
            except Exception as e:
                self.logger.error("Your updating function does not work: " + str(e))
                traceback.print_exc()
                return

            updated_ids = [document["_id"] for document in documents]

            inserted = await self._process_documents(
                dataset_id=updated_dataset_id
                if updated_dataset_id is not None
                else dataset_id,
                documents=updated_documents,
                insert=insert,
                use_json_encoder=use_json_encoder,
            )

            return inserted, updated_ids

        return pull_update_push_subset

    def _validate_cache(
        self, use_cache: bool, dataset_id: str, select_fields: list
    ) -> bool:
        """
        A means to validate the cache by determing that the user desires to
        access the cache and the cache is non-empty.
        """
        num_cached = len(
            list(self._get_cache_path(dataset_id, select_fields).iterdir())
        )
        return use_cache and bool(num_cached)

    def _get_cursor(
        self, dataset_id: str, filters: list, select_fields: list, chunk_size: int
    ) -> str:
        """
        Retrieves the cursor from the backend.
        """
        # The cursor is constrained by its first query. Specifically,
        # if the query from which we get the cursor has a certain page
        # size, further accesses through the cursor will be the same
        # page size. Therefore, in the tasks below, the first task
        # does a proper query of the first {retrieve_chunk_size}
        # documents. Subsequent calls use the cursor to access the
        # next {retrieve_chunk_size} documents {num_requests-1} times.
        # The 0 is there to make the aforementioned statement
        # explicit.
        return self.datasets.documents.get_where(
            dataset_id,
            filters=filters,
            select_fields=select_fields,
            page_size=chunk_size,
            include_vector=False,  # Set to false to for speed
        )["cursor"]

    def _get_task_parameters(
        self,
        valid_cache: bool,
        chunk_size: int,
        dataset_id: str,
        filters: list,
        select_fields: list,
    ) -> Tuple[int, Optional[str]]:
        """
        Retrieves the number of requests (num_requests) and the cursor.
        """
        # num_requests will determine how many requests are sent.
        # Example: Suppose the number of documents is 1254 and
        # retrieve_chunk_size is 100. Then the number of requests would
        # be 1254 // 100, which would be 12, and one, which amounts to
        # 13 requests. This must be defined ahead of time because the
        # cursor predetermines the number of documents to retrieve on
        # the first call. So, The first 12 calls would each get 100
        # documents and call 13 would retrieve the remaining 54.
        if valid_cache:
            print("Accessing cache, disregarding chunk size")
            num_requests = len(
                list(self._get_cache_path(dataset_id, select_fields).iterdir())
            )
            cursor = None
        else:
            # If use_cache is False or num_cached is zero cache will not
            # be accessed.
            num_documents = self.get_number_of_documents(dataset_id, filters)
            num_requests = (num_documents // chunk_size) + 1
            cursor = self._get_cursor(dataset_id, filters, select_fields, chunk_size)

        return num_requests, cursor

    def _get_tasks(
        self,
        valid_cache: bool,
        chunk_size: int,
        dataset_id: str,
        filters: list,
        select_fields: list,
        subset_function,
    ) -> list:
        """
        Gets the list of tasks to be executed in the event loop.
        """
        num_requests, cursor = self._get_task_parameters(
            valid_cache, chunk_size, dataset_id, filters, select_fields
        )
        if valid_cache:
            return [
                subset_function(task_num, 0, cursor) for task_num in range(num_requests)
            ]
        else:
            return [
                subset_function(task_num, chunk_size, None)
                if not task_num
                else subset_function(task_num, 0, cursor)
                for task_num in range(num_requests)
            ]

    def _get_log_file(self, dataset_id) -> str:
        """
        Creates a log file name.
        """
        log_file = "_".join(
            [
                dataset_id,
                str(datetime.now().strftime("%d-%m-%Y-%H-%M-%S")),
                "pull_update_push.log",
            ]
        )
        self.logger.info(f"Created {log_file}")

        return log_file

    def _execute_event_loop(
        self,
        dataset_id: str,
        update_function: Callable[..., list],
        updating_args: dict,
        updated_dataset_id: Optional[str],
        retrieve_chunk_size: Optional[int],
        filters: list,
        select_fields: list,
        use_json_encoder: bool,
        include_vector: bool,
        show_progress_bar: bool,
        insert: bool,
        use_cache: bool,
    ) -> list:
        """
        Primary driver of pull_update_push_async. Retrieves the relevant
        items then executes them in the event loop.
        """
        pull_update_push_subset = self._create_subset_function(
            dataset_id,
            update_function,
            updating_args,
            updated_dataset_id,
            filters,
            select_fields,
            use_json_encoder,
            include_vector,
            insert,
            use_cache,
        )
        valid_cache = self._validate_cache(use_cache, dataset_id, select_fields)
        chunk_size = self._set_chunk_size(
            dataset_id, filters, select_fields, retrieve_chunk_size
        )
        tasks = self._get_tasks(
            valid_cache,
            chunk_size,
            dataset_id,
            filters,
            select_fields,
            pull_update_push_subset,
        )

        threaded_loop = EventLoop(tasks, show_progress_bar)
        threaded_loop.start()
        threaded_loop.execute_tasks()
        threaded_loop.terminate()

        failed_documents = []
        for future in threaded_loop.futures:
            inserted, _ = future.result()
            failed_documents.extend(inserted["failed_documents"])

        return failed_documents


class BatchInsertAsyncClient(BatchInsertAsyncHelpers):
    def pull_update_push_async(
        self,
        dataset_id: str,
        update_function: Callable,
        updating_args: Optional[dict] = None,
        updated_dataset_id: Optional[str] = None,
        log_file: str = None,
        retrieve_chunk_size: Optional[int] = None,
        filters: Optional[list] = None,
        select_fields: Optional[list] = None,
        use_json_encoder: bool = True,
        include_vector: bool = True,
        show_progress_bar: bool = True,
        insert: bool = False,
        use_cache: bool = False,
        log_to_file: bool = True,
    ) -> dict:
        """
        Loops through every document in your collection and applies a function (that is specified by you) to the documents.
        These documents are then uploaded into either an updated collection, or back into the original collection.

        Parameters
        ----------
        dataset_id: str
            The dataset_id of the collection where your original documents are

        update_function: Callable
            A function created by you that converts documents in your original
            collection into the updated documents. The function must contain a
            field which takes in a list of documents from the original
            collection. The output of the function must be a list of updated
            documents.

        updating_args: dict
            Additional arguments to your update_function, if they exist. They
            must be in the format of {'Argument': Value}

        updated_dataset_id: str
            The dataset_id of the collection where your updated documents are
            uploaded into. If 'None', then your original collection will be
            updated.

        log_file: str
            The log file to direct any information or issues that may crop up.
            If no log file is specified, one will automatically be created.

        retrieve_chunk_size: int
            The number of documents that are received from the original
            collection with each loop iteration.

        filters: list
            A list of filters to apply on the retrieval query.

        select_fields: list
            A list of fields to query over.

        use_json_encoder : bool
            If True, automatically convert documents to json encodable format.

        include_vector: bool
            If True, includes vectors in the updating query.

        show_progress_bar: bool
            If True, shows a progress bar.

        insert: bool
            If True, inserts rather than updates an already-existing dataset.

        use_cache: bool
            If True, and if the cache is not empty, uses cached documents.

        log_to_file: bool
            If True, logs errors to a file.
        """
        updating_args = {} if updating_args is None else updating_args
        filters = [] if filters is None else filters
        select_fields = [] if select_fields is None else select_fields

        if not callable(update_function):
            raise TypeError(
                "Your update function needs to be a function! "
                + "Please read the documentation if it is not."
            )

        if not (retrieve_chunk_size is None or isinstance(retrieve_chunk_size, int)):
            raise TypeError(f"{retrieve_chunk_size} must be None or an integer.")

        if log_file is None:
            log_file = self._get_log_file(dataset_id)
        with FileLogger(fn=log_file, verbose=True, log_to_file=log_to_file):
            failed_documents = self._execute_event_loop(
                dataset_id,
                update_function,
                updating_args,
                updated_dataset_id,
                retrieve_chunk_size,
                filters,
                select_fields,
                use_json_encoder,
                include_vector,
                show_progress_bar,
                insert,
                use_cache,
            )
            if failed_documents:
                # This will be picked up by FileLogger
                print("The following documents failed to be updated/inserted:")
                for failed_document in failed_documents:
                    print(f"  * {failed_document}")

        self.logger.success("Pull, update, and push is complete!")

        return {"failed_documents": failed_documents}



from typing import Optional, Tuple, List

from relevanceai._api import APIClient
from relevanceai.client.helpers import Credentials

from relevanceai.utils.decorators.version import added, beta
from relevanceai.utils.logger import FileLogger


class _VectorizeHelper(APIClient):
    def __init__(self, dataset_id: str, **kwargs):
        self.dataset_id = dataset_id
        super().__init__(**kwargs)

    def _check_vector_existence(
        self, schema: dict, fields: List[str], encoder
    ) -> Tuple[list, list, list]:
        """
        This function distinguishes new and existing fields, as well as colect
        metadata.

        Parameters
        ----------
        schema: dict
            The schema of a dataset

        fields: List[str]
            The fields to be distinguished.

        encoder_name:
            The name of the encoder to be applied on the fields.
        """
        new_fields = []
        metadata_additions = []
        existing_vectors = []

        if encoder is not None:
            for field in fields:
                vector_parts = (field, encoder.__name__, "vector_")
                vector = "_".join(vector_parts)
                if vector not in schema:
                    new_fields.append(field)
                    metadata_additions.append(vector_parts)
                else:
                    existing_vectors.append(vector)
                    print(
                        f"Since '{vector}' already exists, its construction "
                        + "will be skipped."
                    )

        return new_fields, metadata_additions, existing_vectors

    def _update_vector_metadata(self, new_vector_metadata: List[tuple]) -> None:
        """
        Given the new vector metdata information, updates the Dataset
        metadata.

        Parameters
        ----------
        new_vector_metadata: List[tuple]
            A list of tuples containing information about the newly created
            vectors.
        """
        updated_metadata = self.datasets.metadata(self.dataset_id)["results"]
        if "_vector_" not in updated_metadata:
            updated_metadata["_vector_"] = {}

        for new_data in new_vector_metadata:
            field, model_name, _ = new_data

            if field not in updated_metadata["_vector_"]:
                updated_metadata["_vector_"][field] = {}

            updated_metadata["_vector_"][field][model_name] = "_".join(new_data)

        self.upsert_metadata(updated_metadata)


class Vectorize(_VectorizeHelper):
    def __init__(
        self,
        credentials: Credentials,
        dataset_id: str,
    ):
        self.dataset_id = dataset_id
        super().__init__(dataset_id=dataset_id, credentials=credentials)

    @beta
    @added(version="1.2.0")
    def vectorize(
        self,
        image_fields: Optional[List[str]] = None,
        text_fields: Optional[List[str]] = None,
        image_encoder=None,
        text_encoder=None,
        log_file: str = "vectorize.logs",
    ) -> dict:
        """
        Parameters
        ----------
        image_fields: List[str]
            A list of image fields to vectorize

        text_fields: List[str]
            A list of text fields to vectorize

        image_encoder
            A deep learning image encoder from the vectorhub library. If no
            encoder is specified, a default encoder (Clip2Vec) is loaded.

        text_encoder
            A deep learning text encoder from the vectorhub library. If no
            encoder is specified, a default encoder (USE2Vec) is loaded.

        Returns
        -------
        dict
            If the vectorization process is successful, this dict contains
            the added vector names. Else, the dict is the request result
            containing error information.

        Example
        -------
        .. code-block::

            from relevanceai import Client
            from vectorhub.encoders.text.sentence_transformers import SentenceTransformer2Vec

            text_model = SentenceTransformer2Vec("all-mpnet-base-v2 ")

            client = Client()

            dataset_id = "sample_dataset_id"
            df = client.Dataset(dataset_id)

            df.vectorize(
                image_fields=["image_field_1", "image_field_2"],
                text_fields=["text_field"],
                text_model=text_model
            )

        """
        schema = self.datasets.schema(self.dataset_id)
        if not image_fields and not text_fields:
            raise ValueError("'image_fields' and 'text_fields' both cannot be empty.")

        image_fields = [] if image_fields is None else image_fields
        text_fields = [] if text_fields is None else text_fields

        fields = image_fields + text_fields

        # we should raise an error here to ensure that if there are any missing fields
        # users know immediately
        foreign_fields = []
        for field in fields:
            if field not in self.schema:
                foreign_fields.append(field)
        else:
            if foreign_fields:
                raise ValueError(
                    f"The following fields are invalid: {', '.join(foreign_fields)}"
                )

        if image_fields and image_encoder is None:
            try:
                with FileLogger(log_file):
                    from vectorhub.bi_encoders.text_image.torch import Clip2Vec

                    image_encoder = Clip2Vec()
                    image_encoder.encode = image_encoder.encode_image
            except ModuleNotFoundError:
                raise ModuleNotFoundError(
                    "Default image encoder not found. "
                    "Please install vectorhub with `python -m pip install "
                    "vectorhub[clip]` to install Clip2Vec."
                )
        if image_fields and not hasattr(image_encoder, "encode_documents"):
            raise AttributeError(
                f"{image_encoder} is missing attribute 'encode_documents'"
            )

        (
            new_image_fields,
            image_metadata,
            existing_image_vectors,
        ) = self._check_vector_existence(schema, image_fields, image_encoder)

        if text_fields and text_encoder is None:
            try:
                with FileLogger(log_file):
                    from vectorhub.encoders.text.tfhub import USE2Vec

                    text_encoder = USE2Vec()
            except ModuleNotFoundError:
                raise ModuleNotFoundError(
                    "Default text encoder not found. "
                    "Please install vectorhub with `python -m pip install "
                    "vectorhub[encoders-text-tfhub]` to install USE2Vec."
                )
        if text_fields and not hasattr(text_encoder, "encode_documents"):
            raise AttributeError(
                f"{text_encoder} is missing attribute 'encode_documents'"
            )

        (
            new_text_fields,
            text_metadata,
            existing_text_vectors,
        ) = self._check_vector_existence(schema, text_fields, text_encoder)

        if new_image_fields or new_text_fields:

            def dual_encoder_function(documents):
                updated_documents = []
                if image_encoder is not None:
                    updated_documents.extend(
                        image_encoder.encode_documents(new_image_fields, documents)
                    )
                if text_encoder is not None:
                    updated_documents.extend(
                        text_encoder.encode_documents(new_text_fields, documents)
                    )

                return updated_documents

            old_schema = schema.keys()

            results = self.pull_update_push(
                self.dataset_id,
                dual_encoder_function,
                select_fields=fields,
                filters=[
                    {
                        "field": field,
                        "filter_type": "exists",
                        "condition": "==",
                        "condition_value": " ",
                        "strict": "must_or",
                    }
                    for field in fields
                ],
            )

            new_schema = schema.keys()

            added_vectors = list(new_schema - old_schema)
        else:
            results = {"failed_documents": []}
            added_vectors = []

        if not results["failed_documents"]:
            if added_vectors:
                print("✅ All documents inserted/edited successfully.")
                self._update_vector_metadata(image_metadata + text_metadata)

            if len(added_vectors) == 1:
                text = "The following vector was added: "
            else:
                text = "The following vectors were added: "
            print(text + ", ".join(added_vectors))

            return {
                "added_vectors": added_vectors,
                "skipped_vectors": existing_image_vectors + existing_text_vectors,
            }
        else:
            print("❗Few errors with vectorizing documents. Please check logs.")
            return results



from typing import Dict, List, Optional
from relevanceai._api import APIClient
from relevanceai.client.helpers import Credentials


class Centroids(APIClient):
    def __init__(self, credentials: Credentials, dataset_id: str):
        self.dataset_id = dataset_id
        super().__init__(credentials)

    def __call__(
        self, vector_fields: list, alias: str, cluster_field: str = "_cluster_"
    ):
        """
        Instaniates Centroids Class which stores centroid information to be called

        Parameters
        ----------
        vector_fields: list
            The vector field where a clustering task was run.
        alias: string
            Alias is used to name a cluster
        cluster_field: string
            Name of clusters in documents

        Example
        --------
        .. code-block
            from relevanceai import Client

            client = Client()

            df = client.Dataset("sample_dataset_id")

            df.get(["sample_id"], include_vector=False)

        """

        self.vector_fields = vector_fields
        self.alias = alias
        self.cluster_field = cluster_field
        self.cluster_doc_field = (
            f"{self.cluster_field}.{self.vector_fields[0]}.{self.alias}"
        )
        return self

    def closest(
        self,
        cluster_ids: Optional[List] = None,
        centroid_vector_fields: Optional[List] = None,
        select_fields: Optional[List] = None,
        approx: int = 0,
        sum_fields: bool = True,
        page_size: int = 1,
        page: int = 1,
        similarity_metric: str = "cosine",
        filters: Optional[List] = None,
        min_score: int = 0,
        include_vector: bool = False,
        include_count: bool = True,
    ):
        """
        List of documents closest from the centre.

        Parameters
        ----------
        cluster_ids: list
            Any of the cluster ids
        centroid_vector_fields: list
            Vector fields stored
        select_fields: list
            Fields to include in the search results, empty array/list means all fields
        approx: int
            Used for approximate search to speed up search. The higher the number, faster the search but potentially less accurate
        sum_fields: bool
            Whether to sum the multiple vectors similarity search score as 1 or seperate
        page_size: int
            Size of each page of results
        page: int
            Page of the results
        similarity_metric: string
            Similarity Metric, choose from ['cosine', 'l1', 'l2', 'dp']
        filters: list
            Query for filtering the search results
        min_score: int
            Minimum score for similarity metric
        include_vectors: bool
            Include vectors in the search results
        include_count: bool
            Include the total count of results in the search results

        Example
        -----------------
        .. code-block::

            from relevanceai import Client
            from relevanceai.ops.clusterer import ClusterOps
            from relevanceai.ops.clusterer.kmeans_clusterer import KMeansModel

            client = Client()

            dataset_id = "sample_dataset_id"
            df = client.Dataset(dataset_id)

            vector_field = "vector_field_"
            n_clusters = 10

            model = KMeansModel(k=n_clusters)

            df.cluster(model=model, alias=f"kmeans-{n_clusters}", vector_fields=[vector_field])

        """
        cluster_ids = [] if cluster_ids is None else cluster_ids
        centroid_vector_fields = (
            [] if centroid_vector_fields is None else centroid_vector_fields
        )
        select_fields = [] if select_fields is None else select_fields
        filters = [] if filters is None else filters

        return self.datasets.cluster.centroids.list_closest_to_center(
            dataset_id=self.dataset_id,
            vector_fields=self.vector_fields,
            alias=self.alias,
            cluster_ids=cluster_ids,
            centroid_vector_fields=centroid_vector_fields,
            select_fields=select_fields,
            approx=approx,
            sum_fields=sum_fields,
            page_size=page_size,
            page=page,
            similarity_metric=similarity_metric,
            filters=filters,
            min_score=min_score,
            include_vector=include_vector,
            include_count=include_count,
        )

    def furthest(
        self,
        cluster_ids: Optional[List] = None,
        centroid_vector_fields: Optional[List] = None,
        select_fields: Optional[List] = None,
        approx: int = 0,
        sum_fields: bool = True,
        page_size: int = 1,
        page: int = 1,
        similarity_metric: str = "cosine",
        filters: Optional[List] = None,
        min_score: int = 0,
        include_vector: bool = False,
        include_count: bool = True,
    ):
        """
        List of documents furthest from the centre.

        Parameters
        ----------
        cluster_ids: list
            Any of the cluster ids
        select_fields: list
            Fields to include in the search results, empty array/list means all fields
        approx: int
            Used for approximate search to speed up search. The higher the number, faster the search but potentially less accurate
        sum_fields: bool
            Whether to sum the multiple vectors similarity search score as 1 or seperate
        page_size: int
            Size of each page of results
        page: int
            Page of the results
        similarity_metric: string
            Similarity Metric, choose from ['cosine', 'l1', 'l2', 'dp']
        filters: list
            Query for filtering the search results
        min_score: int
            Minimum score for similarity metric
        include_vectors: bool
            Include vectors in the search results
        include_count: bool
            Include the total count of results in the search results

        """
        cluster_ids = [] if cluster_ids is None else cluster_ids
        centroid_vector_fields = (
            [] if centroid_vector_fields is None else centroid_vector_fields
        )
        select_fields = [] if select_fields is None else select_fields
        filters = [] if filters is None else filters

        return self.datasets.cluster.centroids.list_furthest_from_center(
            dataset_id=self.dataset_id,
            vector_fields=self.vector_fields,
            alias=self.alias,
            cluster_ids=cluster_ids,
            centroid_vector_fields=centroid_vector_fields,
            select_fields=select_fields,
            approx=approx,
            sum_fields=sum_fields,
            page_size=page_size,
            page=page,
            similarity_metric=similarity_metric,
            filters=filters,
            min_score=min_score,
            include_vector=include_vector,
            include_count=include_count,
        )

    def update(
        self,
        dataset_id: str,
        vector_fields: List[str],
        centroid_vector_fields: List[str],
        alias: str,
        cluster_centers: List[Dict[str, List[float]]],
    ):
        """
        API reference link: https://api.us-east-1.relevance.ai/latest/core/documentation#operation/UpdateClusterCentroids

        Update the centroids contained within your dataset

        Parameters
        ----------

        dataset_id: str
            The name of the dataset

        vector_fields: List[str]
            A lit of the vectors fields in your dataset that have cluster centroids you wish to update

        centroid_vector_fields: List[str]
            ???

        alias: str
            The alias that was used to cluster

        cluster_centers: List[Dict[str: List[float]]]
            A List containing dictionaries of cluster id's to be updated, with their keys being the new centroids
        """

        return self.make_http_request(
            endpoint=f"datasets/{dataset_id}/cluster/centroids/update",
            method="POST",
            parameters={
                "vector_fields": vector_fields,
                "centroid_vector_fields": centroid_vector_fields,
                "alias": alias,
                "cluster_centers": cluster_centers,
            },
        )

    def delete_centroid_by_id(
        self, centroid_id: str, dataset_id: str, vector_field: str, alias: str
    ):
        """
        OLD API reference link: https://api.us-east-1.relevance.ai/latest/documentation#operation/delete_centroids_api_services_cluster_centroids__centroid_id__delete_post

        Delete a centroid by ID

        Parameters
        ----------

        centroid_id: str
            The id of the centroid

        dataset_id: str
            The name of the dataset

        vector_field: str
            The vector_field that contains the cluster id

        alias: str
            The alias that was used to cluster
        """

        return self.make_http_request(
            endpoint=f"services/cluster/centroids/{centroid_id}/delete",
            method="POST",
            parameters={
                "dataset_id": dataset_id,
                "vector_field": vector_field,
                "alias": alias,
            },
        )



"""Base Workflow
"""
from re import I
import traceback
from typing import Callable, Optional
from doc_utils import DocUtils
from uuid import uuid4


class Workflow(DocUtils):
    """
    Base Workflow. A workflow is useful for measuring what you did with a dataset.
    By adding an alias, you allow it to continue running even when it errors.
    """

    def __init__(
        self, func: Callable, workflow_alias: str, notes: Optional[str] = None
    ):
        """ """
        self.func = func
        self.workflow_alias = workflow_alias
        self.notes = notes

    def fit_dataset(
        self,
        dataset,
        input_field: str,
        output_field: str,
        filters: Optional[list] = None,
        log_to_file: bool = True,
        chunksize: int = 20,
        chunk_field: Optional[str] = None,
    ):
        """
        Fit on dataset
        """
        filters = [] if filters is None else filters
        self.dataset = dataset
        exist_filters = [
            {
                "field": input_field,
                "filter_type": "exists",
                "condition": "==",
                "condition_value": " ",
            },
            {
                "field": output_field,
                "filter_type": "exists",
                "condition": "!=",
                "condition_value": " ",
            },
        ]
        filters += exist_filters

        if chunk_field is not None:
            if chunk_field not in output_field:
                output_field = chunk_field + "." + output_field

            if chunk_field not in input_field:
                input_field = chunk_field + "." + input_field

        def update_func(doc):
            if chunk_field:
                try:
                    self.run_function_across_chunks(
                        function=self.func,
                        chunk_field=chunk_field,
                        field=input_field,
                        output_field=output_field,
                        doc=doc,
                    )
                except Exception as e:
                    traceback.print_exc()
            else:
                try:
                    value = self.get_field(input_field, doc)
                    self.set_field(output_field, doc, self.func(value))
                except Exception as e:
                    traceback.print_exc()
            return doc

        # Store this workflow inside the dataset's metadata
        results = dataset.apply(
            update_func,
            select_fields=[input_field],
            filters=filters,
            log_to_file=log_to_file,
            retrieve_chunksize=chunksize,
        )
        self._store_workflow_to_metadata(input_field, output_field)
        return results

    def _store_workflow_to_metadata(
        self, input_field, output_field, run_id: str = None
    ):
        if run_id is None:
            run_id = uuid4().__str__()
        workflow_metadata = {
            "input_field": input_field,
            "output_field": output_field,
            "workflow_alias": self.workflow_alias,
            "run_id": run_id,
        }
        if self.notes is not None:
            workflow_metadata["notes"] = self.notes
        metadata = self.dataset.metadata
        if "workflows" not in metadata:
            metadata["workflows"] = []
        workflows = metadata["workflows"]
        workflows.append(workflow_metadata)
        self.dataset.metadata["workflows"] = workflows
